# Standard libraries
# from common import dtype_MAPPING
import json
import os
from datetime import datetime
from pathlib import Path

from loguru import logger


class JSONWriter:
    """Class that handles writing simulation results to json
    files.
    """

    def __init__(self, log_dir: Path, column_length: int):
        """Constructor for json writer

        Args:
            log_dir (Path): output path (dirctory)
            column_length (int): number of data columns
        """
        self.column_length = column_length
        self.log_dir = log_dir
        self.chunk_rows = 1
        self.handles = {}
        self.dfs = {}
        self.step = 0
        # Create arrow writer for each object type

    def write(self, currenttime: datetime, powerflow_output: dict):
        """Writes the status of assets at a particular timestep to a json file.

        Args:
            currenttime (datetime): simulator time step
            powerflow_output (dict): simulation results
        """
        # Iterate through each object type

        for obj_type in powerflow_output:
            fpath = os.path.join(self.log_dir, f"{obj_type}.json")
            if self.step == 0:
                f = open(fpath, "w")
                f.close()
                self.handles[obj_type] = open(fpath, "a")
            data = powerflow_output[obj_type]
            if obj_type not in self.dfs:
                self.dfs[obj_type] = {str(currenttime): data}
            else:
                if self.dfs[obj_type] is None:
                    self.dfs[obj_type] = {str(currenttime): data}
                else:
                    self.dfs[obj_type][currenttime] = data
            if self.step % self.chunk_rows == self.chunk_rows - 1:
                try:
                    json.dump(self.dfs[obj_type], self.handles[obj_type], indent=4)
                    self.handles[obj_type].flush()
                    self.dfs[obj_type] = None
                except Exception as E:
                    logger.warning(f"Unable to write property {obj_type} to file: {E!s}")

        self.step += 1



from enum import Enum, IntEnum


class ProjectFolders(str, Enum):
    "Defines valid the project folder"
    CASESTUDY = "case_study"
    EXPORTS = "exports"
    GISDATA = "gis_data"
    LOGS = "logs"
    PROFILES = "profiles"


class BulkWriteModes(str, Enum):
    "Supported bulk writers"
    CSV = "csv"
    PKL = "pkl"


class StreamedWriteModes(str, Enum):
    "Supported stream writers"
    H5 = "h5"


class SimulationModes(str, Enum):
    "Valid PyPSSE simulation modes"
    PCM = "PCM"
    SNAP = "Snap"
    STATIC = "Steady-state"
    DYNAMIC = "Dynamic"


class HelicsCoreTypes(str, Enum):
    "HELICS core types"
    ZMQ = "zmq"


class WritableModelTypes(str, Enum):
    "Writable data ty[es]"
    LOAD = "Load"
    PLANT = "Plant"
    MACHINE = "Machine"
    GENERATOR = "Induction_machine"


class ModelTypes(str, Enum):
    "Supported asset tpyes in PyPSSE"
    BUSES = "Buses"
    BRANCHES = "Branches"
    LOADS = "Loads"
    GENERATORS = "Induction_generators"
    MACHINES = "Machines"
    FIXED_SHUNTS = "Fixed_shunts"
    SWITCHED_SHUNTS = "Switched_shunts"
    TRANSFORMERS = "Transformers"
    AREAS = "Areas"
    ZONES = "Zones"
    DC_LINES = "DCtransmissionlines"
    STATIONS = "Stations"


class ModelProperties(str, Enum):
    "Support model properties"
    PU = "PU"
    FREQ = "FREQ"
    ANGLE = "ANGLE"
    ANGLED = "ANGLED"


class LoggingLevels(IntEnum):
    "logging level setting options"
    NOTSET = 0
    DEBUG = 10
    INFO = 20
    WARN = 30
    ERROR = 40
    CRITICAL = 50


class SubscriptionFileRequiredColumns(str, Enum):
    "Subscription file requirements"
    bus_subsystem_id = "bus_subsystem_id"
    element_type = "element_type"
    element_id = "element_id"
    element_property = "element_property"
    sub_tag = "sub_tag"
    scaler = "scaler"
    bus = "bus"


class BusProperties(str, Enum):
    "Valid bus properties"

    BASE = "BASE"
    FREQ = "FREQ"
    PU = "PU"
    KV = "KV"
    ANGLE = "ANGLE"
    ANGLED = "ANGLED"
    NVLMHI = "NVLMHI"
    NVLMLO = "NVLMLO"
    EVLMHI = "EVLMHI"
    EVLMLO = "EVLMLO"


class AreaProperties(str, Enum):
    "Valid area properties"

    LOAD = "LOAD"
    LOADID = "LOADID"
    LGDN = "LGDN"
    LDGNLD = "LDGNLD"
    GEN = "GEN"
    AREANAME = "AREANAME"
    AREANUMBER = "AREANUMBER"


class ZoneProperties(str, Enum):
    "Valid zone properties"

    LOAD = "LOAD"
    LOADID = "LOADID"
    LGDN = "LGDN"
    LDGNLD = "LDGNLD"
    GEN = "GEN"
    ZONENAME = "ZONENAME"
    ZONENUMBER = "ZONENUMBER"


class StationProperties(str, Enum):
    "Valid station properties"

    SUBNAME = "SUBNAME"
    SUBNUMBER = "SUBNUMBER"
    BUSES = "BUSES"
    GENERATORS = "GENERATORS"
    TRANSFORMERS = "TRANSFORMERS"
    NOMKV = "NOMKV"
    LOADMW = "LOADMW"
    GENMW = "GENMW"


class DCLineProperties(str, Enum):
    "Valid DC line properties"

    DCLINENAME = "DCLINENAME"
    MDC = "MDC"
    RECT = "RECT"
    INV = "INV"
    METER = "METER"
    NBR = "NBR"
    NBI = "NBI"
    ICR = "ICR"
    ICI = "ICI"
    NDR = "NDR"
    NDI = "NDI"


class LoadProperties(str, Enum):
    "Valid load properties"

    MVA = "MVA"
    IL = "IL"
    YL = "YL"
    TOTAL = "TOTAL"
    YNEG = "YNEG"
    YZERO = "YZERO"
    FmA = "FmA"
    FmB = "FmB"
    FmC = "FmC"
    FmD = "FmD"
    Fel = "Fel"
    PFel = "PFel"
    TD = "TD"
    TC = "TC"


class FixedShuntProperties(str, Enum):
    "Valid fixed shunt properties"

    ACT = "ACT"
    O_ACT = "O_ACT"
    NOM = "NOM"
    O_NOM = "O_NOM"
    PQZERO = "PQZERO"
    PQZ = "PQZ"
    O_PQZ = "O_PQZ"


class SwitchedShuntProperties(str, Enum):
    "Valid switched shunt properties"

    VSWHI = "VSWHI"
    VSWLO = "VSWLO"
    RMPCT = "RMPCT"
    BINIT = "BINIT"
    O_BINIT = "O_BINIT"


class TransformerProperties(str, Enum):
    "Valid transformer properties"

    RATIO = "RATIO"
    RATIO2 = "RATIO2"
    ANGLE = "ANGLE"
    RMAX = "RMAX"
    RMIN = "RMIN"
    VMAX = "VMAX"
    VMIN = "VMIN"
    STEP = "STEP"
    CR = "CR"
    CX = "CX"
    CNXANG = "CNXANG"
    SBASE1 = "SBASE1"
    NOMV1 = "NOMV1"
    NOMV2 = "NOMV2"
    GMAGNT = "GMAGNT"
    BMAGNT = "BMAGNT"
    RG1 = "RG1"
    XG1 = "XG1"
    R01 = "R01"
    X01 = "X01"
    RG2 = "RG2"
    XG2 = "XG2"
    R02 = "R02"
    X02 = "X02"
    RNUTRL = "RNUTRL"
    XNUTRL = "XNUTRL"
    RX1_2 = "RX1-2"
    RX2_3 = "RX2-3"
    RX3_1 = "RX3-1"
    YMAGNT = "YMAGNT"
    ZG1 = "ZG1"
    Z01 = "Z01"
    ZG2 = "ZG2"
    Z02 = "Z02"
    ZG3 = "ZG3"
    Z03 = "Z03"
    ZNUTRL = "ZNUTRL"


class BranchProperties(str, Enum):
    "Valid branch properties"

    RATEn = "RATEn"
    RATEA = "RATEA"
    RATEB = "RATEB"
    RATEC = "RATEC"
    RATE = "RATE"
    LENGTH = "LENGTH"
    CHARG = "CHARG"
    CHARGZ = "CHARGZ"
    FRACT1 = "FRACT1"
    FRACT2 = "FRACT2"
    FRACT3 = "FRACT3"
    FRACT4 = "FRACT4"
    RX = "RX"
    ISHNT = "ISHNT"
    JSHNT = "JSHNT"
    RXZ = "RXZ"
    ISHNTZ = "ISHNTZ"
    JSHNTZ = "JSHNTZ"
    LOSSES = "LOSSES"
    O_LOSSES = "O_LOSSES"
    MVA = "MVA"
    AMPS = "AMPS"
    PUCUR = "PUCUR"
    CURANG = "CURANG"
    P = "P"
    O_P = "O_P"
    Q = "Q"
    O_Q = "O_Q"
    PLOS = "PLOS"
    O_PLOS = "O_PLOS"
    QLOS = "QLOS"
    O_QLOS = "O_QLOS"


class InductionGeneratorProperties(str, Enum):
    "Valid induction generator properties"

    MBASE = "MBASE"
    RATEKV = "RATEKV"
    PSET = "PSET"
    RA = "RA"
    XA = "XA"
    R1 = "R1"
    X1 = "X1"
    R2 = "R2"
    X2 = "X2"
    X3 = "X3"
    E1 = "E1"
    SE1 = "SE1"
    E2 = "E2"
    SE2 = "SE2"
    IA1 = "IA1"
    IA2 = "IA2"
    XAMULT = "XAMULT"
    TRQA = "TRQA"
    TRQB = "TRQB"
    TRQD = "TRQD"
    TRQE = "TRQE"
    H = "H"
    IRATIO = "IRATIO"
    ROVERX = "ROVERX"
    RZERO = "RZERO"
    XZERO = "XZERO"
    RGRND = "RGRND"
    XGRND = "XGRND"
    P = "P"
    O_P = "O_P"
    Q = "Q"
    O_Q = "O_Q"
    MVA = "MVA"
    O_MVA = "O_MVA"
    SLIP = "SLIP"
    ZA = "ZA"
    Z1 = "Z1"
    Z2 = "Z2"
    ZZERO = "ZZERO"
    ZGRND = "ZGRND"
    PQ = "PQ"
    O_PQ = "O_PQ"


class MachinesProperties(str, Enum):
    "Valid machine properties"

    QMAX = "QMAX"
    O_QMAX = "O_QMAX"
    QMIN = "QMIN"
    O_QMIN = "O_QMIN"
    PMAX = "PMAX"
    O_PMAX = "O_PMAX"
    PMIN = "PMIN"
    O_PMIN = "O_PMIN"
    MBASE = "MBASE"
    MVA = "MVA"
    O_MVA = "O_MVA"
    P = "P"
    O_P = "O_P"
    Q = "Q"
    O_Q = "O_Q"
    PERCENT = "PERCENT"
    GENTAP = "GENTAP"
    VSCHED = "VSCHED"
    WPF = "WPF"
    RMPCT = "RMPCT"
    RPOS = "RPOS"
    XSUBTR = "XSUBTR"
    XTRANS = "XTRANS"
    XSYNCH = "XSYNCH"
    PQ = "PQ"
    O_PQ = "O_PQ"
    ZSORCE = "ZSORCE"
    XTRAN = "XTRAN"
    ZPOS = "ZPOS"
    ZNEG = "ZNEG"
    ZZERO = "ZZERO"
    ZGRND = "ZGRND"


Channels = {
    "ANGLE = false machine relative rotor angle (degrees).": False,
    "PELEC = false machine electrical power (pu on SBASE).": False,
    "QELEC = false machine reactive power.": False,
    "ETERM = false machine terminal voltage (pu).": False,
    "EFD = false generator main field voltage (pu).": False,
    "PMECH = false turbine mechanical power (pu on MBASE).": False,
    "SPEED = false machine speed deviation from nominal (pu).": False,
    "XADIFD = false machine field current (pu).": False,
    "ECOMP = false voltage regulator compensated voltage (pu).": False,
    "VOTHSG = false stabilizer output signal (pu).": False,
    "VREF = false voltage regulator voltage setpoint (pu).": False,
    "BSFREQ = false bus pu frequency deviations.": False,
    "VOLT = false bus pu voltages (complex).": False,
    "voltage and angle.": False,
    "flow (P).": False,
    "flow (P and Q).": False,
    "flow (MVA).": False,
    "apparent impedance (R and X).": False,
    "ITERM.": False,
    "machine apparent impedance.": False,
    "VUEL = false minimum excitation limiter output signal (pu).": False,
    "VOEL = false maximum excitation limiter output signal (pu).": False,
    "PLOAD.": False,
    "QLOAD.": False,
    "GREF = false turbine governor reference.": False,
    "LCREF = false turbine load control reference.": False,
    "WVLCTY = false wind velocity (m/s).": False,
    "WTRBSP = false wind turbine rotor speed deviation (pu).": False,
    "WPITCH = false pitch angle (degrees).": False,
    "WAEROT = false aerodynamic torque (pu on MBASE).": False,
    "WROTRV = false rotor voltage (pu on MBASE).": False,
    "WROTRI = false rotor current (pu on MBASE).": False,
    "WPCMND = false active power command from wind control (pu on MBASE).": False,
    "WQCMND = false reactive power command from wind control (pu on MBASE).": False,
}


class ExportModes(str, Enum):
    "Valid export modes"

    CSV = "csv"
    H5 = "h5"


class ChannelTypes(str, Enum):
    "Valid channel types"

    BUSES = "buses"
    LOADS = "loads"
    MACHINES = "machines"


class UseModes(str, Enum):
    "Valid use modes"

    REGEX = "regex"
    LIST = "list"
    ALL = "all"


class ApiCommands(str, Enum):
    RUN_SIMUALTION = "run_simulation"
    STATUS = "status"
    SOLVE_STEP = "run_step"
    OPEN_CASE = "open_case"
    CLOSE_CASE = "close_case"
    UDPATE_MODEL = "update_model"
    UPDATE_PARAMETERS = "update_settings"
    QUERY_ALL = "query_all"
    QUERY_BY_ID = "query_by_asset"
    QUERY_BY_PPTY = "query_by_ppty"
    QUERY_ASSET_LIST = "query_asset_list"


class SimulationStatus(str, Enum):
    NOT_INITIALIZED = "Instance not initialized"
    STARTING_INSTANCE = "Starting PyPSSE instance"
    INITIALIZATION_COMPLETE = "PyPSSE initialization complete"
    RUNNING_SIMULATION = "Running simulation"
    SIMULATION_COMPLETE = "Simulation complete"
    STARTING_RESULT_EXPORT = "Starting exports"
    RESULT_EXPORT_COMPLETE = "Export coplete"

class PSSE_VERSIONS(str, Enum):
    PSSE34 = "psse34"
    PSSE35 = "psse35"
    PSSE36 = "psse36"


# Standard imports

import json
import logging
from http import HTTPStatus
from inspect import signature
from multiprocessing import current_process
from queue import Empty
from types import FunctionType
from typing import Union

from pypsse.api.common import BASE_PROJECT_PATH
from pypsse.common import MAPPED_CLASS_NAMES, SIMULATION_SETTINGS_FILENAME
from pypsse.enumerations import ApiCommands, SimulationStatus
from pypsse.models import ApiAssetQuery, ApiPssePostRequest, ApiPsseReply, ApiWebSocketRequest, SimulationSettings
from pypsse.simulator import Simulator
from pypsse.utils.utils import load_settings

logger = logging.getLogger(__name__)
# Make sure to change the PSSE path


class SimulatorUtils:
    def open_case(
        self,
        parameters: Union[ApiPssePostRequest, dict],
    ) -> str:
        """run the case file.

        Args:
            parameters (Union[ApiPssePostRequest, dict]): should contain valid project name

        Returns:
            str: exception returned as string
        """
        if isinstance(parameters, dict):
            parameters = ApiPssePostRequest(**parameters)

        project_path = BASE_PROJECT_PATH / parameters.project_name
        assert project_path.exists(), f"Project path {project_path!s} does not exist"
        simulation_filepath = project_path / SIMULATION_SETTINGS_FILENAME
        try:
            settings = load_settings(simulation_filepath)
            self.logger.info(SimulationStatus.STARTING_INSTANCE.value)
            self.psse_obj = Simulator(settings)
            self.logger.info(SimulationStatus.INITIALIZATION_COMPLETE.value)
            self.time = self.psse_obj.settings.simulation.start_time
        except Exception as e:
            return str(e)
        return

    def close_case(self):
        """method can be used to close a currently open project"""

        self.psse_obj.PSSE.pssehalt_2()
        del self.psse_obj
        logger.info(f"PSSE case {self.uuid} closed.")

    def status(self) -> SimulationStatus:
        """method returns simulation state

        Returns:
            SimulationStatus: simulation state
        """

        if self.psse_obj is None:
            return SimulationStatus.NOT_INITIALIZED.value
        return self.psse_obj._status

    def run_simulation(self):
        """use method to run the simulation"""
        self.psse_obj.run()

    def run_step(self) -> SimulationStatus:
        """use the method to run a single simulation step

        Returns:
            SimulationStatus: simulation results
        """
        t = (self.time - self.psse_obj.settings.simulation.start_time).total_seconds()
        self.current_result = self.psse_obj.step(t)
        self.time += self.psse_obj.settings.simulation.simulation_step_resolution
        self._restructure_results()
        return self.psse_obj._status

    def export_result(self):
        """use method to export simulation results"""
        self.psse_obj.PSSE.pssehalt_2()
        if not self.psse_obj.export_settings.export_results_using_channels:
            self.psse_obj.results.export_results()
        else:
            self.psse_obj.sim.export()

    def update_model(self, parameters):
        # TODO: import to implement @Aadil
        ...

    def update_settings(self, simulation_settings: SimulationSettings):
        """method updates simulation settings

        Args:
            simulation_settings (SimulationSettings): simulation settings
        """
        self.psse_obj.settings = self.psse_obj.settings.model_copy(update=simulation_settings)

    def query_asset_list(self) -> dict:
        """use method to query list of assets in loaded model

        Returns:
            dict: mapping of asset type to assets
        """
        results = {}
        for asset_type, asset_info in self.results_by_id.items():
            if asset_type not in results:
                results[asset_type] = list(asset_info)
        return results

    def query_all(self) -> dict:
        """use method to query all results

        Returns:
            dict: all simulation results
        """
        results = self.results_by_ppty
        return results

    def query_by_asset(self, parameters: Union[ApiAssetQuery, dict]) -> dict:
        """use method to filter results for an asset type

        Args:
            parameters (Union[ApiAssetQuery, dict]): asset type to be queried

        Returns:
            dict: filtered simulation results
        """

        if isinstance(parameters, dict):
            parameters = ApiAssetQuery(**parameters)

        assert hasattr(parameters, "asset_type") and hasattr(
            parameters, "asset_id"
        ), "Reqired keys: asset_type, asset_id"
        inv_map = {v: k for k, v in MAPPED_CLASS_NAMES.items()}
        asset_type = inv_map[parameters.asset_type]
        return self.results_by_id[asset_type][parameters.asset_id]

    def query_by_ppty(self, parameters: Union[ApiAssetQuery, dict]) -> dict:
        """use method to filter results for a property

        Args:
            parameters (Union[ApiAssetQuery, dict]): asset properties to be queried

        Returns:
            dict: filtered simualtion results
        """
        if isinstance(parameters, dict):
            parameters = ApiAssetQuery(**parameters)

        assert hasattr(parameters, "asset_type") and hasattr(
            parameters, "asset_property"
        ), "Reqired keys: asset_type, asset_property"
        inv_map = {v: k for k, v in MAPPED_CLASS_NAMES.items()}
        asset_type = inv_map[parameters.asset_type]
        return self.results_by_ppty[asset_type][parameters.asset_property]

    def _validate_methods(self):
        """validates api commands match available methods"""
        for command in ApiCommands:
            assert hasattr(self, command.value)

    def _restructure_results(self):
        """retturns restructured simulation results"""

        inv_map = {v: k for k, v in MAPPED_CLASS_NAMES.items()}
        self.results_by_id = {}
        self.results_by_ppty = {}
        for info, val_dict in self.current_result.items():
            asset_type, asset_property = info.split("_")
            new_asset_type = inv_map[asset_type]
            for asset_id, value in val_dict.items():
                if new_asset_type not in self.results_by_id:
                    self.results_by_id[new_asset_type] = {}
                    self.results_by_ppty[new_asset_type] = {}
                if asset_id not in self.results_by_id[new_asset_type]:
                    self.results_by_id[new_asset_type][asset_id] = {}
                if asset_property not in self.results_by_ppty[new_asset_type]:
                    self.results_by_ppty[new_asset_type][asset_property] = {}

                if isinstance(value, complex):
                    val = [value.real, value.imag]
                elif isinstance(value, list):
                    val = []
                    for v in value:
                        val.append([v.real, v.imag])
                else:
                    val = value

                val_struct = {"value": val, "dtype": str(value.__class__.__name__)}

                self.results_by_id[new_asset_type][asset_id][asset_property] = val_struct
                self.results_by_ppty[new_asset_type][asset_property][asset_id] = val_struct


class SimulatorWebSocket(SimulatorUtils):
    """_summary_

    Args:
        SimulatorUtils (__type__): methods implemented for the simulation handler
    """

    def __init__(
        self,
        shutdown_event=None,
        to_psse_queue=None,
        from_psse_queue=None,
    ):
        """Implmenmts the handler for the web socket implementation

        Args:
            shutdown_event (_type_, optional): event used to shutdown the simulation. Defaults to None.
            to_psse_queue (_type_, optional): Queue use to push to the simulation handler. Defaults to None.
            from_psse_queue (_type_, optional): Queue used by simulator to push back to the API. Defaults to None.
        """
        super().__init__()
        self._validate_methods()
        self.uuid = current_process().name

        # Setting up logger for the process
        # qh = logging.handlers.QueueHandler(logging_queue)
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.DEBUG)
        # self.logger.addHandler(qh)
        self.logger.info("%s - psse uuid init time", self.uuid)

        self.shutdownevent = shutdown_event
        self.to_psse_queue = to_psse_queue
        self.from_psse_queue = from_psse_queue

        # PyPSSE object responsible for communicating to PSSE API
        self.psse_obj = None
        self.time_counter = 0
        self.multipart, self.multipart_data = {}, {}
        # TODO: I don't know if we need to check at this point
        # whether we can import psse35 and psspy

        self._add_message_to_queue(
            ApiPsseReply(status=HTTPStatus.OK, message="sucessfully initialized.", uuid=self.uuid)
        )
        self.run()

    def _get_arguments(self, command_str: str) -> [FunctionType, list]:
        """method to match method to provided parameters

        Args:
            command_str (str): methods name

        Returns:
            [FunctionType, list]: reference to mapped method, list of accepted method parameters
        """

        # Check if the command specified by the
        # user is defined as method in class
        if not hasattr(self, command_str):
            self._add_message_to_queue(
                ApiPsseReply(
                    status=HTTPStatus.BAD_REQUEST, message=f"Command {command_str} does not exist.", uuid=self.uuid
                )
            )
            return
        command = getattr(self, command_str)
        return command, list(signature(command).parameters)

    def _add_message_to_queue(self, model_instance: ApiPsseReply):
        """used to push message to out going queue

        Args:
            model_instance (ApiPsseReply): message
        """
        # Enums are not JSON serializable
        # so need to first convert to json and load into pythin dict
        self.from_psse_queue.put(json.loads(model_instance.model_dump_json()))

    def _get_task(self):
        """Internal method to parse task."""

        task = self.to_psse_queue.get()

        try:
            if task == "END" or task is None:
                return task
            return ApiWebSocketRequest(**json.loads(task))  # json.loads(task)

        # pylint: disable=broad-exception-caught
        except Exception as err:
            msg = f"Error parsing the inputs as dictionary: {err!s} > input: {task}"
            self.logger.error(msg)
            self._add_message_to_queue(
                ApiPsseReply(status=HTTPStatus.INTERNAL_SERVER_ERROR, message=msg, uuid=self.uuid)
            )

    def run(self) -> None:
        """execute the method for the web socket interface.
        runs indefinitey until shutdown by setting
        the event or END command is sent

        Returns:
            _type_: message pushed to the from_psse  queue
        """

        self.logger.info("PSSE simulation starting")

        while not self.shutdownevent.is_set():
            try:
                # Fetch the task from queue break
                task = self._get_task()

                if task is None:
                    continue

                if task == "END":
                    break

                func, attrs = self._get_arguments(task.command)

                try:
                    if "parameters" in attrs:
                        command_return = func(parameters=task.parameters)
                    else:
                        command_return = func()

                    result = ApiPsseReply(status=HTTPStatus.OK, message=command_return, uuid=self.uuid)

                except Exception as e:
                    result = ApiPsseReply(status=HTTPStatus.INTERNAL_SERVER_ERROR, message=str(e), uuid=self.uuid)
                self._add_message_to_queue(result)

            except Empty:
                continue

            except (KeyboardInterrupt, SystemExit):
                break

        self.logger.info("%s - finishing PSSE simulation", self.uuid)
        msg = "Closed PSSE instance"
        self._add_message_to_queue(ApiPsseReply(status=HTTPStatus.OK, message=msg, uuid=self.uuid))


class SimulatorAPI(SimulatorUtils):
    def __init__(
        self, shutdown_event=None, to_psse_queue=None, from_psse_queue=None, params: ApiPssePostRequest = None
    ):
        """Implmenmts the handler for the web socket implementation

        Args:
            shutdown_event (_type_, optional): event used to shutdown the simulation. Defaults to None.
            to_psse_queue (_type_, optional): Queue use to push to the simulation handler. Defaults to None.
            from_psse_queue (_type_, optional): Queue used by simulator to push back to the API. Defaults to None
            params (ApiPssePostRequest, optional): _description_. Defaults to None.
        """
        super().__init__()

        self._validate_methods()

        self.uuid = current_process().name
        logger.info(f"{self.uuid} - psse uuid init time")
        self.shutdownevent = shutdown_event
        self.to_psse_queue = to_psse_queue
        self.from_psse_queue = from_psse_queue

        project_path = BASE_PROJECT_PATH / params.project_name
        assert project_path.exists(), f"Project path {project_path!s} does not exist"
        simulation_filepath = project_path / SIMULATION_SETTINGS_FILENAME
        try:
            settings = load_settings(simulation_filepath)
            self.psse_obj = Simulator(settings)
            logger.info(f"{self.uuid} - psse dispatched")
            result = ApiPsseReply(
                status=HTTPStatus.OK, message=f"Simulation {self.uuid} successfully initiallized.", uuid=self.uuid
            )
            if from_psse_queue is not None:
                self.from_psse_queue.put(result)
            self.time = self.psse_obj.settings.simulation.start_time
            self.run()
        except Exception as e:
            result = ApiPsseReply(status=HTTPStatus.INTERNAL_SERVER_ERROR, message=str(e), uuid=self.uuid)
            logger.error(result.model_dump_json())

    def run(self):
        """execute the method for the REST api interface.
        runs indefinitey until shutdown by setting
        the event or END command is sent

        Returns:
            _type_: message pushed to the from_psse  queue
        """
        logger.info("{} - Simulation starting")
        while not self.shutdownevent.is_set():
            try:
                task = self.to_psse_queue.get()
                if task == "END":
                    break
                else:
                    try:
                        task = json.loads(task)
                    except Exception as e:
                        logger.error(str(e))
                        self.from_psse_queue.put(ApiPsseReply(status=HTTPStatus.INTERNAL_SERVER_ERROR, message=str(e)))

                    command_name = task.pop("command")

                    if hasattr(self, command_name):
                        command = getattr(self, command_name)
                        try:
                            if "parameters" in list(signature(command).parameters):
                                command_return = command(parameters=task["parameters"])
                            else:
                                command_return = command()
                            print("command_return: ", command_return)
                            result = ApiPsseReply(status=HTTPStatus.OK, message=json.dumps(command_return))
                            print("ApiPsseReply: ", result)
                        except Exception as e:
                            result = ApiPsseReply(status=HTTPStatus.INTERNAL_SERVER_ERROR, message=str(e))

                    else:
                        result = ApiPsseReply(
                            status=HTTPStatus.UNPROCESSABLE_ENTITY, message=f"Command {command_name} does not exist."
                        )

                self.from_psse_queue.put(result)
            except Empty:
                continue
            except (KeyboardInterrupt, SystemExit):
                break
        msg = f"{self.uuid} - finishing simulation"
        logger.info(msg)
        result = ApiPsseReply(status=HTTPStatus.OK, message=msg)
        logger.info(msg)
        self.from_psse_queue.put(result)



# Standard imports
import datetime
import os

from loguru import logger

# Internal imports
from pypsse.modes.abstract_mode import AbstractMode
from pypsse.utils.dc2ac.dc_ac_algorithm import DC2ACconverter


class ProductionCostModel(AbstractMode):
    def __init__(self, psse, dyntools, settings, export_settings, subsystem_buses, raw_data):
        super().__init__(psse, dyntools, settings, export_settings, subsystem_buses, raw_data)
        self.time = datetime.datetime.strptime(settings["Simulation"]["Start time"], "%m/%d/%Y %H:%M:%S").astimezone(
            None
        )
        self._StartTime = datetime.datetime.strptime(
            settings["Simulation"]["Start time"], "%m/%d/%Y %H:%M:%S"
        ).astimezone(None)
        self.incTime = settings["Simulation"]["Step resolution (sec)"]
        self.convergence_helper = DC2ACconverter(psse, self, settings, raw_data)

    def init(self, bussubsystems):
        super().init(bussubsystems)
        self.initialization_complete = True

    def step(self):
        ierr = self.PSSE.fnsl()
        assert ierr == 0, f"Error code: {ierr}"
        solved_flag = self.PSSE.solved()
        # check if powerflow completed successfully
        if solved_flag == 0:
            self.time = self.time + datetime.timedelta(seconds=self.incTime)
            return 0
        else:
            self.convergence_helper.run()
            if not self.convergence_helper.has_converged:
                logger.error(
                    f"Error code {solved_flag} returned from PSSE while running powerflow, please follow \
                                PSSE doumentation to know more about error"
                )
                return 1
            else:
                self.time = self.time + datetime.timedelta(seconds=self.incTime)
                return 0

    def reload(self):
        pass

    def resolve_step(self):
        ierr = self.PSSE.fnsl()
        assert ierr == 0, f"Error code: {ierr}"
        solved_flag = self.PSSE.solved()
        if solved_flag > 0:
            self.convergence_helper.run()
            if not self.convergence_helper.has_converged:
                msg = f"Error code {solved_flag} returned from PSSE while running powerflow, please follow \
                                            PSSE doumentation to know more about error"
                raise Exception(msg)

    def get_time(self):
        return self.time

    def get_total_seconds(self):
        return (self.time - self._StartTime).total_seconds()

    def get_step_size_cec(self):
        return self.settings["Simulation"]["Step resolution (sec)"]

    def export(self):
        logger.debug("Starting export process. Can take a few minutes for large files")
        excelpath = os.path.join(self.export_path, self.settings["Excel file"])
        achnf = self.dyntools.CHNF(self.outx_path)
        achnf.xlsout(channels="", show=False, xlsfile=excelpath, outfile="", sheet="Sheet1", overwritesheet=True)
        logger.debug("{} export to {}".format(self.settings["Simulation"]["Excel file"], self.export_path))



import json
from pathlib import Path

import numpy as np
import openmdao.api as om
import toml
from loguru import logger

from pypsse.models import MdaoProblem, SimulationSettings
from pypsse.simulator import Simulator


class PSSE:
    "The class defines the PSSE interface to OpenMDAO"
    model_loaded = False

    def load_model(self, settings_file_path: Path):
        """Load the PyPSSE model

        Args:
            settings_file_path (Path): path to simulation setting file
        """

        settings = toml.load(settings_file_path)
        settings["simulation"]["project_path"] = settings_file_path.parent
        settings = SimulationSettings(**settings)
        self.psse_obj = Simulator(settings)
        self.assets = self.psse_obj.raw_data
        self.time_counter = 0
        self.psse_obj.init()
        self.model_loaded = True

    def _build_inputs(self) -> dict:
        """builds mdao inputs

        Returns:
            dict: mapping of input variables to values
        """
        inputs_dict = {}
        for input_model in self.probelm.inputs:
            self.psse_obj.sim.update_object(
                dtype=input_model.asset_type.value,
                bus=input_model.asset_bus,
                element_id=input_model.asset_id,
                values=input_model.attributes,
            )
            for var, val in input_model.attributes.items():
                tag = f"{input_model.asset_type.value}_{input_model.asset_id}_{input_model.asset_bus}_{var}"
                inputs_dict[tag] = [val]
        logger.info("MDAO inputs built sucessfully")
        self.solve_step()
        return inputs_dict

    def _list_inputs(self) -> list:
        """list subproblem input variables

        Returns:
            list: list of input variables
        """
        return list(self._psse_inputs.keys())

    def _build_outputs(self, output: dict = None) -> dict:
        """Updates outputs if dict provided and returns output vaiable dict

        Args:
            output (dict, optional): mapping of variables to value. Defaults to None.

        Returns:
            dict: mapping of variables to value
        """
        outputs = json.loads(self.probelm.outputs.model_dump_json())
        buses = outputs["buses"]
        quantities = outputs["quantities"]

        results = self.psse_obj.sim.read_subsystems(subsystem_buses=buses, quantities=quantities)
        if output is None:
            output = {}
        for obj_ppty, values in results.items():
            asset_type, asset_property = obj_ppty.split("_")
            for obj_id, value in values.items():
                if isinstance(obj_id, int):
                    bus_id = obj_id
                    object_id = None
                else:
                    obj_id = obj_id.replace(" ", "")
                    bus_id, object_id = obj_ppty.split("_")
                if object_id:
                    tag = f"{asset_type}_{object_id}_{bus_id}_{asset_property}"
                else:
                    tag = f"{asset_type}_{bus_id}_{asset_property}"

                if isinstance(value, complex):
                    output[f"{tag}_real"] = [value.real]
                    output[f"{tag}_imag"] = [value.imag]
                else:
                    output[f"{tag}"] = [value]
        logger.info("MDAO outputs built sucessfully")
        return output

    def _update_inputs(self, inputs: dict):
        """updates input variables

        Args:
            inputs (dict): mapping of input variables to values
        """
        attr_keys = {}
        for _input in inputs:
            k, attr = _input.rsplit("_", 1)
            if k not in attr_keys:
                attr_keys[k] = {}
            attr_keys[k][attr] = inputs[_input][0]

        for info, attrs in attr_keys.items():
            asset_type, asset_id, asset_bus_id = info.split("_")
            self.psse_obj.sim.update_object(dtype=asset_type, bus=int(asset_bus_id), element_id=asset_id, values=attrs)
        logger.info("MDAO inputs updated sucessfully")

    def solve_step(self):
        """Solves for the current time set and incremetn in time"""

        self.psse_obj.inc_time = False
        self.current_result = self.psse_obj.step(self.time_counter)

    def export_result(self):
        """Updates results in the result container"""

        if not self.psse_obj.export_settings.export_results_using_channels:
            self.psse_obj.results.export_results()
        else:
            self.psse_obj.sim.export()
        logger.info("Result sucessfully exported")

    def close_case(self):
        """Closes the loaded model in PyPSSE"""

        self.psse_obj.psse.pssehalt_2()
        del self.psse_obj
        logger.info("PSSE case closed.")

    def read_problem_data(self, problem_file: Path):
        """reads the mdao config file

        Args:
            problem_file (Path): path to mdao problem defination toml file
        """
        data = toml.load(problem_file)
        self.probelm = MdaoProblem(**data)
        logger.info("MDAO probelm parameters read. Building inputs and outputs.")

    def __del__(self):
        self.export_result()
        self.close_case()


class PypsseMdaoModel(om.ExplicitComponent, PSSE):
    "Expicit OpenMDAO component"

    def __init__(self, settings_file_path: Path, problem_file: Path):
        """initializes the optimization problem

        Args:
            settings_file_path (Path): pypsse simulation settings file
            problem_file (Path): mdao problem defination toml file
        """

        self.read_problem_data(problem_file)
        self.case = self.load_model(settings_file_path)
        super().__init__()
        self._psse_inputs = self._build_inputs()
        for var, val in self._psse_inputs.items():
            self.add_input(var, val=val)

        self._psse_outputs = self._build_outputs()
        for var, val in self._psse_outputs.items():
            self.add_output(var, val=val)

    def setup(self):
        """Sets up the optimization problem"""

    def setup_partials(self):
        """Sets up the problem partial derivatives"""
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs: object, outputs: object):
        """Sets up the compute method for openmdao

        Args:
            inputs (object): openmdao input objects
            outputs (object): openmdao output objects
        """

        self._psse_inputs = inputs
        self._psse_outputs = outputs
        self._update_inputs(inputs)
        self.solve_step()
        self._build_outputs(outputs)

    def __repr__(self) -> str:
        """overrides defaulr 'print' behavior

        Returns:
            str: problem defination summary
        """
        input_str = ""
        for _input, val in self._psse_inputs.items():
            input_str += f" {_input} - {val}\n"

        output_str = ""
        for output, val in self._psse_outputs.items():
            output_str += f" {output} - {val}\n"

        msg = f"""Inputs:\n{input_str}\nOutputs:\n{output_str}"""
        return msg



# Standard libraries
# from common import dtype_MAPPING
import os
from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd


class CSVWriter:
    """Class that handles writing simulation results to csv
    files.
    """

    def __init__(self, log_dir: Path, column_length: int):
        """Constructor for csv writer

        Args:
            log_dir (Path): output path (dirctory)
            column_length (int): number of data columns
        """

        self.column_length = column_length
        self.log_dir = log_dir
        self.timestamps = []
        self.chunkRows = 1
        self.dfs = {}
        self.step = 0
        # Create arrow writer for each object type

    def write(self, currenttime: datetime, powerflow_output: dict):
        """Writes the status of assets at a particular timestep to a csv file.

        Args:
            currenttime (datetime): simulator time step
            powerflow_output (dict): simulation results
        """

        # Iterate through each object type
        self.timestamps.append(currenttime)
        for obj_type in powerflow_output:
            data = powerflow_output[obj_type]
            if obj_type not in self.dfs:
                self.dfs[obj_type] = [data]
            else:
                if self.dfs[obj_type] is None:
                    self.dfs[obj_type] = [data]
                else:
                    self.dfs[obj_type].append(data)

            if self.step % self.chunkRows == self.chunkRows - 1:
                fpath = os.path.join(self.log_dir, f"{obj_type}.csv")
                self.dfs[obj_type] = pd.DataFrame(self.dfs[obj_type], index=self.timestamps)
                self.dfs[obj_type].to_csv(fpath, mode="a")

                self.dfs[obj_type] = None
            self.Timestamp[self.step - 1] = np.string_(str(currenttime))
        self.step += 1



from pathlib import Path

import toml

from pypsse.models import SimulationSettings


def load_settings(file_path: Path, example_path=None) -> SimulationSettings:
    """utility function to load the simulation file

    Args:
        file_path (Path): simulation file path
        example_path (_type_, optional): project path. Defaults to None.

    Returns:
        SimulationSettings: simulation settings
    """
    settings = toml.load(file_path)
    if example_path:
        settings["simulation"]["project_path"] = example_path
    settings = SimulationSettings(**settings)
    return settings



# Standard libraries
# from common import dtype_MAPPING
import os
from datetime import datetime
from pathlib import Path

import h5py
import numpy as np
import pandas as pd
from loguru import logger

from pypsse.common import DEFAULT_RESULTS_FILENAME


class HDF5Writer:
    """Class that handles writing simulation results to hdf5 files."""

    def __init__(self, log_dir: Path, column_length: int):
        """Constructor for hdf5 writer

        Args:
            log_dir (Path): output path (dirctory)
            column_length (int): number of data columns
        """

        self.log_dir = log_dir
        self.store = h5py.File(os.path.join(log_dir, DEFAULT_RESULTS_FILENAME), "w")
        self.store_groups = {}
        self.store_datasets = {}
        self.row = {}
        self.column_length = column_length

        self.chunkRows = 1
        self.step = 0
        self.dfs = {}
        self.convergence = self.store.create_dataset(
            "Convergence",
            shape=(self.column_length,),
            maxshape=(None,),
            chunks=True,
            compression="gzip",
            compression_opts=4,
            shuffle=True,
            dtype=np.int16,
        )
        self.Timestamp = self.store.create_dataset(
            "Time stamp",
            shape=(self.column_length,),
            maxshape=(None,),
            chunks=True,
            compression="gzip",
            compression_opts=4,
            shuffle=True,
            dtype="S30",
        )
        # Create arrow writer for each object type

    def write(self, currenttime: datetime, powerflow_output: dict, convergence: int):
        """Writes the status of assets at a particular timestep to a hdf5 file.

        Args:
            currenttime (datetime): simulator time step
            powerflow_output (dict): simulation results
            convergence (int): simulation convergence status
        """
        # Iterate through each object type
        for obj_type in powerflow_output:
            data = pd.DataFrame(powerflow_output[obj_type], index=[self.step])
            data = data.fillna(0)

            if obj_type not in self.row:
                self.row[obj_type] = 0
                self.store_groups[obj_type] = self.store.create_group(obj_type)
                self.store_datasets[obj_type] = {}
                for col_name in powerflow_output[obj_type].keys():
                    dtype = data[col_name].dtype
                    if dtype == object:
                        dtype = "S30"

                    self.store_datasets[obj_type][col_name] = self.store_groups[obj_type].create_dataset(
                        str(col_name),
                        shape=(self.column_length,),
                        maxshape=(None,),
                        chunks=True,
                        compression="gzip",
                        compression_opts=4,
                        shuffle=True,
                        dtype=dtype,
                    )

            if obj_type not in self.dfs:
                self.dfs[obj_type] = data
            else:
                if self.dfs[obj_type] is None:
                    self.dfs[obj_type] = data
                else:
                    self.dfs[obj_type] = self.dfs[obj_type].append(data, ignore_index=True)

            if self.step % self.chunkRows == self.chunkRows - 1:
                si = int(self.step / self.chunkRows) * self.chunkRows
                ei = si + self.chunkRows
                for col_name in powerflow_output[obj_type].keys():
                    r = self.store_datasets[obj_type][col_name].shape[0]
                    if ei >= r:
                        self.store_datasets[obj_type][col_name].resize((ei,))
                    self.store_datasets[obj_type][col_name][si:ei] = self.dfs[obj_type][col_name]
                self.dfs[obj_type] = None
            if self.step >= len(self.Timestamp):
                self.Timestamp.resize((len(self.Timestamp) + 1,))
                self.convergence.resize((len(self.convergence) + 1,))
            self.Timestamp[self.step - 1] = np.string_(str(currenttime))
            self.convergence[self.step - 1] = convergence
            # Add object status data to a DataFrame
            self.store.flush()
        self.step += 1

    def __del__(self):
        try:
            k = next(self.dfs.keys())
            length = len(self.dfs[k])
            if self.dfs[k] is not None:
                for obj_type in self.dfs.keys():
                    for col_name in self.dfs[k].columns:
                        self.store_datasets[obj_type][col_name][self.column_length - length :] = self.dfs[obj_type][
                            col_name
                        ]

            self.store.flush()
            self.store.close()
        except Exception as e:
            logger.info(str(e))



"""Mapping of all channels available in PSS/E
"""


channel_map = {
    "ANGLE, machine relative rotor angle (degrees).": 1,
    "PELEC, machine electrical power (pu on SBASE).": 2,
    "QELEC, machine reactive power.": 3,
    "ETERM, machine terminal voltage (pu).": 4,
    "EFD, generator main field voltage (pu).": 5,
    "PMECH, turbine mechanical power (pu on MBASE).": 6,
    "SPEED, machine speed deviation from nominal (pu).": 7,
    "XADIFD, machine field current (pu).": 8,
    "ECOMP, voltage regulator compensated voltage (pu).": 9,
    "VOTHSG, stabilizer output signal (pu).": 10,
    "VREF, voltage regulator voltage setpoint (pu).": 11,
    "BSFREQ, bus pu frequency deviations.": 12,
    "VOLT, bus pu voltages (complex).": 13,
    "voltage and angle.": 14,
    "flow (P).": 15,
    "flow (P and Q).": 16,
    "flow (MVA).": 17,
    "apparent impedance (R and X).": 18,
    "ITERM.": 21,
    "machine apparent impedance.": 22,
    "VUEL, minimum excitation limiter output signal (pu).": 23,
    "VOEL, maximum excitation limiter output signal (pu).": 24,
    "PLOAD.": 25,
    "QLOAD.": 26,
    "GREF, turbine governor reference.": 27,
    "LCREF, turbine load control reference.": 28,
    "WVLCTY, wind velocity (m/s).": 29,
    "WTRBSP, wind turbine rotor speed deviation (pu).": 30,
    "WPITCH, pitch angle (degrees).": 31,
    "WAEROT, aerodynamic torque (pu on MBASE).": 32,
    "WROTRV, rotor voltage (pu on MBASE).": 33,
    "WROTRI, rotor current (pu on MBASE).": 34,
    "WPCMND, active power command from wind control (pu on MBASE).": 35,
    "WQCMND, reactive power command from wind control (pu on MBASE).": 36,
}



# Standard imports

import numpy as np
from loguru import logger

from pypsse.common import CASESTUDY_FOLDER, VALUE_UPDATE_BOUND
from pypsse.enumerations import ModelTypes, WritableModelTypes
from pypsse.models import ExportSettings, SimulationModes, SimulationSettings
from pypsse.modes.constants import converter


class AbstractMode:
    def __init__(
        self,
        psse,
        dyntools,
        settings: SimulationSettings,
        export_settings: ExportSettings,
        subsystem_buses,
        raw_data,
    ):
        self.psse = psse
        self.raw_data = raw_data
        self.bus_freq_channels = {}

        from psspy import _f, _i, _o, _s

        self._i = _i
        self._f = _f
        self._s = _s
        self._o = _o
        self.bus_freq_channels = {}

        self.sub_buses = subsystem_buses
        self.dyntools = dyntools
        self.settings = settings
        self.export_settings = export_settings
        self.func_options = {
            ModelTypes.BUSES.value: {
                "busdat": ["BASE", "PU", "KV", "ANGLE", "ANGLED", "NVLMHI", "NVLMLO", "EVLMHI", "EVLMLO"],
                "busdt2": ["TOTAL", "FX_TOTAL", "SC_TOTAL", "YS", "YSW"],  # requires string2 input
                "busint": ["STATION", "TYPE", "AREA", "ZONE", "OWNER", "DUMMY"],
                "arenam": ["AREANAME"],
                "stadat": ["LATI", "LONG"],
                "gendat": ["GENPOWER"],
                "notona": ["NAME"],
                "busexs": ["STATUS"],
                "busnofunc": ["NUMBER", "ISLOADBUS"],
                "frequency": ["FREQ"],
            },
            ModelTypes.STATIONS.value: {
                "stanofunc": ["SUBNAME", "SUBNUMBER", "BUSES", "GENERATORS", "TRANSFORMERS", "NOMKV", "LOADMW", "GENMW"]
            },
            ModelTypes.AREAS.value: {
                "arenofunc": ["AREANAME", "AREANUMBER"],
                "ardat": ["LOAD", "LOADLD", "LDGN", "LDGNLD", "GEN"],
            },
            ModelTypes.ZONES.value: {
                "zonnofunc": ["ZONENAME", "ZONENUMBER"],
                "zndat": ["LOAD", "LOADID", "LDGN", "LDGNLD", "GEN", "LOSS"],
            },
            ModelTypes.DC_LINES.value: {
                "dctnofunc": ["DCLINENAME"],
                "dc2int_2": ["MDC", "RECT", "INV", "METER", "NBR", "NBI", "ICR", "ICI", "NDR", "NDI"],
            },
            ModelTypes.BRANCHES.value: {
                "brndat": [
                    "RATEn",
                    "RATEA",
                    "RATEB",
                    "RATEC",
                    "RATE",
                    "LENGTH",
                    "CHARG",
                    "CHARGZ",
                    "FRACT1",
                    "FRACT2",
                    "FRACT3",
                    "FRACT4",
                ],
                "brndt2": ["RX", "ISHNT", "JSHNT", "RXZ", "ISHNTZ", "JSHNTZ", "LOSSES", "O_LOSSES", "RX"],
                "brnmsc": [
                    "MVA",
                    "AMPS",
                    "PUCUR",
                    "CURANG",
                    "P",
                    "O_P",
                    "Q",
                    "O_Q",
                    "PLOS",
                    "O_PLOS",
                    "QLOS",
                    "O_QLOS",
                    "PCTRTA",
                ],
                "brnint": [
                    "STATUS",
                    "METER",
                    "NMETR",
                    "OWNERS",
                    "OWN1",
                    "OWN2",
                    "OWN3",
                    "OWN4",
                    "STATION_I",
                    "STATION_J",
                    "SECTION_I",
                    "SECTION_J",
                    "NODE_I",
                    "NODE_J",
                    "SCTYPE",
                ],
                "brnnofunc": [
                    "FROMBUSNUM",
                    "TOBUSNUM",
                    "FROMBUSNAME",
                    "TOBUSNAME",
                    "CIRCUIT",
                    "SUBNUMBERTO",
                    "SUBNUMBERFROM",
                    "NOMKVFROM",
                    "NOMKVTO",
                    "BY",
                ],
            },
            ModelTypes.GENERATORS.value: {
                "inddt1": [
                    "MBASE",
                    "RATEKV",
                    "PSET",
                    "RA",
                    "XA",
                    "R1",
                    "X1",
                    "R2",
                    "X2",
                    "X3",
                    "E1",
                    "SE1",
                    "E2",
                    "SE2",
                    "IA1",
                    "IA2",
                    "XAMULT",
                    "TRQA",
                    "TRQB",
                    "TRQD",
                    "TRQE",
                    "H",
                    "IRATIO",
                    "ROVERX",
                    "RZERO",
                    "XZERO",
                    "RGRND",
                    "XGRND",
                    "P",
                    "O_P",
                    "Q",
                    "O_Q",
                    "MVA",
                    "O_MVA",
                    "SLIP",
                ],
                "inddt2": ["ZA", "Z1", "Z2", "ZZERO", "ZGRND", "PQ", "O_PQ"],
                "indnofunc": ["INDID", "BUSNUM", "BUSNAME"],
            },
            ModelTypes.LOADS.value: {
                "loddt2": ["MVA", "IL", "YL", "TOTAL", "YNEG", "YZERO"],  # required string 2 input
                "lodnofunc": ["LOADID", "BUSNUM", "BUSNAME"],
                "lodint": ["STATION", "SECTION", "STATUS", "AREA", "ZONE", "OWNER", "SCALE", "CGR"],
            },
            ModelTypes.MACHINES.value: {
                "macdat": [
                    "QMAX",
                    "O_QMAX",
                    "QMIN",
                    "O_QMIN",
                    "PMAX",
                    "O_PMAX",
                    "PMIN",
                    "O_PMIN",
                    "MBASE",
                    "MVA",
                    "O_MVA",
                    "P",
                    "O_P",
                    "Q",
                    "O_Q",
                    "PERCENT",
                    "GENTAP",
                    "VSCHED",
                    "WPF",
                    "RMPCT",
                    "RPOS",
                    "XSUBTR",
                    "XTRANS",
                    "XSYNCH",
                ],
                "macdt2": ["PQ", "O_PQ", "ZSORCE", "XTRAN", "ZPOS", "ZNEG", "ZZERO", "ZGRND"],
                "macnofunc": ["MACID", "BUSNUM", "BUSNAME", "SUBNUMBER", "SUBLATITUDE", "SUBLONGITUDE", "AREANUMBER"],
                "macint": ["STATION", "SECTION", "STATUS", "IREG", "NREG", "OWNERS", "WMOD", "PERCENT", "CZG"],
            },
            ModelTypes.FIXED_SHUNTS.value: {
                "fxsdt2": ["ACT", "O_ACT", "NOM", "O_NOM", "PQZERO", "PQZ", "O_PQZ"],
                "fxsnofunc": ["FXSHID", "BUSNUM", "BUSNAME"],
            },
            ModelTypes.SWITCHED_SHUNTS.value: {
                "swsdt1": ["VSWHI", "VSWLO", "RMPCT", "BINIT", "O_BINIT"],
                "swsnofunc": ["BUSNUM", "BUSNAME"],
            },
            ModelTypes.TRANSFORMERS.value: {
                "xfrdat": [
                    "RATIO",
                    "RATIO2",
                    "ANGLE",
                    "RMAX",
                    "RMIN",
                    "VMAX",
                    "VMIN",
                    "STEP",
                    "CR",
                    "CX",
                    "CNXANG",
                    "SBASE1",
                    "NOMV1",
                    "NOMV2",
                    "NOMV2",
                    "GMAGNT",
                    "BMAGNT",
                    "RG1",
                    "XG1",
                    "R01",
                    "X01",
                    "RG2",
                    "XG2",
                    "R02",
                    "X02",
                    "RNUTRL",
                    "XNUTRL",
                ],
                "tr3dt2": ["RX1-2", "RX2-3", "RX3-1", "YMAGNT", "ZG1", "Z01", "ZG2", "Z02", "ZG3", "Z03", "ZNUTRL"],
                "trfnofunc": [
                    "FROMBUSNUM_3WDG",
                    "FROMBUSNUM_2WDG",
                    "TOBUSNUM_3WDG",
                    "TOBUSNUM_2WDG",
                    "TOBUS2NUM_3WDG",
                    "FROMBUSNAME_3WDG",
                    "FROMBUSNAME_2WDG",
                    "TOBUSNAME_3WDG",
                    "TOBUSNAME_2WDG",
                    "TOBUS2NAME_3WDG",
                    "CIRCUIT_2WDG",
                    "CIRCUIT_3WDG",
                ],
            },
        }
        self.initialization_complete = False

    def save_model(self):
        export_path = self.settings.simulation.project_path / CASESTUDY_FOLDER

        i = 0
        file_path = export_path / f"modified_steady_state_{i}.sav"
        while file_path.exists():
            file_path = export_path / f"modified_steady_state_{i}.sav"
            i += 1

        savfile = export_path / f"modified_steady_state_{i}.sav"
        rawfile = export_path / f"modified_steady_state_{i}.raw"
        self.psse.save(str(savfile))
        logger.info(f"file saved - {savfile}")
        self.psse.rawd_2(0, 1, [1, 1, 1, 0, 0, 0, 0], 0, str(rawfile))
        logger.info(f"file saved - {rawfile}")
        if self.settings.simulation.simulation_mode in [SimulationModes.DYNAMIC, SimulationModes.SNAP]:
            snpfile = export_path / f"modified_dynamic_{i}.snp"
            self.psse.snap([-1, -1, -1, -1, -1], str(snpfile))
            logger.info(f"file saved - {snpfile}")
            return [savfile, rawfile, snpfile]

        else:
            return [savfile, rawfile]

    def init(self, bus_subsystems=None):
        assert bus_subsystems is not None
        if self.settings.log.disable_psse_logging:
            self.disable_logging()

    def step(self):
        return

    def resolve_step(self):
        return

    def export(self):
        try:
            logger.info("Starting export process. Can take a few minutes for large files")
            achnf = self.dyntools.CHNF(str(self.settings.export.outx_file))
            achnf.xlsout(
                channels="",
                show=False,
                xlsfile=str(self.settings.export.excel_file),
                outfile="",
                sheet="Sheet1",
                overwritesheet=True,
            )
            logger.debug(f"Exported {self.settings.export.excel_file}")
        except Exception as e:
            raise e

    def load_user_defined_models(self):
        for mdl in self.settings.simulation.user_models:
            self.psse.addmodellibrary(str(mdl))
            logger.debug(f"User defined library added: {mdl}")

    def load_setup_files(self):
        for f in self.settings.simulation.setup_files:
            ierr = self.psse.runrspnsfile(str(f))
            if ierr:
                msg = f'Error running setup file "{f}"'
                raise Exception(msg)
            else:
                logger.debug(f"Setup file {f} sucessfully run")

    def disable_logging(self):
        self.psse.progress_output(islct=6, filarg="", options=[0, 0])
        self.psse.prompt_output(islct=6, filarg="", options=[0, 0])
        self.psse.report_output(islct=6, filarg="", options=[0, 0])

    def setup_channels(self):
        from pypsse import channel_map

        for channel, add in self.export_settings["Channels"].items():
            if add:
                channel_id = channel_map.channel_map[channel]
                logger.debug(f'"{channel}" added to the channel file')
                self.psse.chsb(0, 1, [-1, -1, -1, 1, channel_id, 0])

    def get_area_numbers(self, subsystem_buses):
        area_numbers = []
        for b in subsystem_buses:
            ierr, val = self.psse.busint(int(b), "AREA")

            if val:
                area_numbers.append(val)
        return list(set(area_numbers))

    def get_substation_numbers(self, subsystem_buses):
        substation_numbers = []
        for b in subsystem_buses:
            irr, val = self.psse.busint(int(b), "STATION")
            if val:
                substation_numbers.append(val)
        return list(set(substation_numbers))

    def get_zone_numbers(self, subsystem_buses):
        zone_numbers = []
        for b in subsystem_buses:
            irr, val = self.psse.busint(int(b), "ZONE")
            if val:
                zone_numbers.append(val)
        return list(set(zone_numbers))

    def get_dctr_line_names(self, subsystem_buses):
        dc_lines = []
        ierr = self.psse.ini2dc()
        if ierr:
            logger.info("No DC line in the model.")
        else:
            ierr, dc_line_name = self.psse.nxtmdc()
            while dc_line_name is not None:
                ierr, rectbus = self.psse.dc2int_2(dc_line_name, "RECT")
                ierr, invbus = self.psse.dc2int_2(dc_line_name, "INV")
                if rectbus in subsystem_buses and invbus in subsystem_buses:
                    dc_lines.append(dc_line_name)

                ierr, dc_line_name = self.psse.nxt2dc()
        return list(set(dc_lines))

    def get_a_list_of_buses_in_substation(self, sub_number, subsystem_buses):
        bus_numbers = []
        for b in subsystem_buses:
            ierr, val = self.psse.busint(int(b), "STATION")

            if val == sub_number:
                bus_numbers.append(int(b))
        return list(set(bus_numbers))

    def get_a_list_of_nomkv_in_substation(self, sub_number, subsystem_buses):
        nom_kvs = []
        for b in subsystem_buses:
            ierr, val = self.psse.busint(int(b), "STATION")

            if val == sub_number:
                ierr, val = self.psse.busdat(int(b), "BASE")

                nom_kvs.append(val)
        return list(set(nom_kvs))

    def get_a_loadmw_in_substation(self, sub_number, subsystem_buses):
        load_mw = 0
        for b in subsystem_buses:
            ierr, val = self.psse.busint(int(b), "STATION")

            if val == sub_number:
                ierr = self.psse.inilod(int(b))

                ierr, load_id = self.psse.nxtlod(int(b))

                while ierr == 0:
                    ierr, val = self.psse.loddt2(int(b), load_id, "TOTAL", "ACT")

                    if isinstance(val, complex):
                        load_mw += val.real
                    ierr, load_id = self.psse.nxtlod(int(b))

        return load_mw

    def get_a_genmw_in_substation(self, sub_number, subsystem_buses):
        gen_mw = 0
        for b in subsystem_buses:
            ierr, val = self.psse.busint(int(b), "STATION")

            if val == sub_number:
                ierr, val = self.psse.gendat(int(b))

                if isinstance(val, complex):
                    gen_mw += val.real
        return gen_mw

    def get_a_list_of_generators_in_substation(self, sub_number, subsystem_buses):
        generators = []
        for b in subsystem_buses:
            ierr, val = self.psse.busint(int(b), "STATION")

            if val == sub_number:
                ierr = self.psse.inimac(int(b))

                ierr, mach_id = self.psse.nxtmac(int(b))

                while ierr == 0:
                    gen_name = f"{b}_{mach_id}"
                    ierr, mach_id = self.psse.nxtmac(int(b))

                    generators.append(gen_name)

        return list(set(generators))

    def get_a_list_of_transformers_in_substation(self, sub_number, subsystem_buses):
        transformers = []
        for b in subsystem_buses:
            ierr, val = self.psse.busint(int(b), "STATION")

            if val == sub_number:
                # Get two winding transformers
                ierr = self.psse.inibrx(int(b), 1)

                ierr, b1, ickt = self.psse.nxtbrn(int(b))

                ickt_string = str(ickt)
                while ierr == 0:
                    ierr, val = self.psse.xfrdat(int(b), int(b1), ickt, "RATIO")

                    if ierr == 0:
                        t_name = f"{b}_{b1}_{ickt_string}"
                        transformers.append(t_name)
                    ierr, b1, ickt = self.psse.nxtbrn(int(b))

                # Get three winding transformers
                ierr = self.psse.inibrx(int(b), 1)

                ierr, b1, b2, ickt = self.psse.nxtbrn3(int(b))

                ickt_string = str(ickt)
                while ierr == 0:
                    ierr, val = self.psse.tr3dt2(int(b), int(b1), int(b2), ickt, "Z01")

                    if ierr == 0:
                        t_name = f"{b!s}_{b1!s}_{b2!s}_{ickt_string}"
                        transformers.append(t_name)
                    ierr, b1, b2, ickt = self.psse.nxtbrn3(int(b))

        return list(set(transformers))

    def check_for_loadbus(self, b):
        ierr = self.psse.inilod(int(b))

        if ierr == 0:
            ierr = self.psse.inimac(int(b))
            if ierr:
                return 1

        return 0

    @converter
    def read_subsystems(self, quantities, subsystem_buses, ext_string2_info=None, mapping_dict=None):
        if ext_string2_info is None:
            ext_string2_info = {}
        if mapping_dict is None:
            mapping_dict = {}

        results = {}
        area_numbers = self.get_area_numbers(subsystem_buses)
        zone_numbers = self.get_zone_numbers(subsystem_buses)
        dctr_lines = self.get_dctr_line_names(subsystem_buses)
        substation_numbers = self.get_substation_numbers(subsystem_buses)

        for class_name, var_list in quantities.items():
            if class_name in self.func_options:
                funcs = self.func_options[class_name]
                for id_, v in enumerate(var_list):
                    for func_name, settinsgs in funcs.items():
                        if v in settinsgs:
                            q = f"{class_name}_{v}"

                            if len(mapping_dict) != 0:
                                if class_name in mapping_dict:
                                    new_v = mapping_dict[class_name][id_]
                                    q = f"{class_name}_{new_v}"

                            if class_name == ModelTypes.AREAS.value:
                                for arr_num in area_numbers:
                                    if func_name in ["arenofunc"]:
                                        if v == "AREANUMBER":
                                            results = self.add_result(results, q, int(arr_num), arr_num)
                                        elif v == "AREANAME":
                                            ierr, val = self.psse.arenam(int(arr_num))

                                            results = self.add_result(results, q, val, arr_num)

                            elif class_name == ModelTypes.STATIONS.value:
                                for sub_num in substation_numbers:
                                    if func_name in ["stanofunc"]:
                                        if v == "SUBNUMBER":
                                            results = self.add_result(results, q, int(sub_num), sub_num)
                                        elif v == "SUBNAME":
                                            ierr, val = self.psse.staname(int(sub_num))

                                            results = self.add_result(results, q, val, sub_num)
                                        elif v == "BUSES":
                                            val = self.get_a_list_of_buses_in_substation(sub_num, subsystem_buses)
                                            results = self.add_result(results, q, val, sub_num)
                                        elif v == "GENERATORS":
                                            val = self.get_a_list_of_generators_in_substation(sub_num, subsystem_buses)
                                            results = self.add_result(results, q, val, sub_num)
                                        elif v == "TRANSFORMERS":
                                            val = self.get_a_list_of_transformers_in_substation(
                                                sub_num, subsystem_buses
                                            )
                                            results = self.add_result(results, q, val, sub_num)
                                        elif v == "NOMKV":
                                            val = self.get_a_list_of_nomkv_in_substation(sub_num, subsystem_buses)
                                            results = self.add_result(results, q, max(val), sub_num)
                                        elif v == "LOADMW":
                                            val = self.get_a_loadmw_in_substation(sub_num, subsystem_buses)
                                            results = self.add_result(results, q, val, sub_num)
                                        elif v == "GENMW":
                                            val = self.get_a_genmw_in_substation(sub_num, subsystem_buses)
                                            results = self.add_result(results, q, val, sub_num)

                            elif class_name == ModelTypes.ZONES.value:
                                for zn_num in zone_numbers:
                                    if func_name in ["zonnofunc"]:
                                        if v == "ZONENUMBER":
                                            results = self.add_result(results, q, int(zn_num), zn_num)
                                        elif v == "ZONENAME":
                                            ierr, val = self.psse.zonnam(int(zn_num))

                                            results = self.add_result(results, q, val, zn_num)

                            elif class_name == ModelTypes.DC_LINES.value:
                                for dcline in dctr_lines:
                                    if func_name == "dctnofunc":
                                        if v == "DCLINENAME":
                                            results = self.add_result(results, q, dcline, dcline)
                                    elif func_name == "dc2int_2":
                                        ierr, val = getattr(self.psse, func_name)(dcline, v)

                                        results = self.add_result(results, q, val, dcline)

                            else:
                                for b in subsystem_buses:
                                    if func_name in [
                                        "busdat",
                                        "busdt2",
                                        "busint",
                                        "arenam",
                                        "notona",
                                        "busexs",
                                        "gendat",
                                        "busnofunc",
                                        "frequency",
                                    ]:
                                        if func_name == "busnofunc":
                                            if v == "NUMBER":
                                                results = self.add_result(results, q, int(b), b)
                                            if v == "ISLOADBUS":
                                                val = self.check_for_loadbus(b)
                                                results = self.add_result(results, q, val, b)

                                        if func_name == "frequency":
                                            if v == "FREQ":
                                                if b in self.bus_freq_channels:
                                                    ierr, val = self.psse.chnval(self.bus_freq_channels[b])

                                                    if not val:
                                                        val = np.NaN
                                                    results = self.add_result(results, q, val * 60.0, int(b))
                                                else:
                                                    results = self.add_result(results, q, 0, int(b))

                                        if func_name in ["busdat", "busint"]:
                                            # if v == 'FREQ':
                                            ierr, val = getattr(self.psse, func_name)(int(b), v)

                                            results = self.add_result(results, q, val, b)

                                        elif func_name == "busdt2":
                                            string2 = "ACT"
                                            if class_name in ext_string2_info:
                                                if v in ext_string2_info[class_name]:
                                                    string2 = ext_string2_info[class_name][v]

                                            ierr, val = getattr(self.psse, func_name)(int(b), v, string2)

                                            results = self.add_result(results, q, val, b)

                                        elif func_name in ["notona", "gendat"]:
                                            ierr, val = getattr(self.psse, func_name)(int(b))

                                            results = self.add_result(results, q, val, b)

                                        elif func_name == "busexs":
                                            val = getattr(self.psse, func_name)(int(b))
                                            val = not val  # PSSE is weird, 0 means bus exist so had to negate
                                            results = self.add_result(results, q, val, b)

                                        elif func_name == "arenam":
                                            ierr, val = self.psse.busint(int(b), "AREA")

                                            if val:
                                                irr, val = getattr(self.psse, func_name)(val)
                                            results = self.add_result(results, q, val, b)

                                    elif func_name == "stadat":
                                        ierr, val = self.psse.busint(int(b), "STATION")

                                        if val:
                                            irr, val = getattr(self.psse, func_name)(val, v)
                                        results = self.add_result(results, q, val, b)

                                    elif func_name in ["inddt1", "inddt2", "indnofunc"]:
                                        ierr = self.psse.iniind(int(b))
                                        if ierr:
                                            logger.info("No induction machine in the model")
                                        else:
                                            ierr, ind_id = self.psse.nxtind(int(b))
                                            while ind_id is not None:
                                                if func_name == "indnofunc":
                                                    if v in ["BUSNUM", "INDID"]:
                                                        val = {"INDID": ind_id, "BUSNUM": int(b)}[v]
                                                        results = self.add_result(results, q, val, f"{ind_id}_{b}")
                                                    elif v == "BUSNAME":
                                                        irr, val = self.psse.notona(int(b))
                                                        results = self.add_result(results, q, val, f"{ind_id}_{b}")
                                                else:
                                                    ierr, val = getattr(self.psse, func_name)(int(b), ind_id, v)
                                                    results = self.add_result(results, q, val, f"{ind_id}_{b}")
                                                ierr, ind_id = self.psse.nxtind(int(b))

                                    elif func_name in ["loddt2", "lodnofunc", "lodint"]:
                                        ierr = self.psse.inilod(int(b))

                                        ierr, load_id = self.psse.nxtlod(int(b))

                                        while load_id is not None:
                                            if func_name == "lodnofunc":
                                                if v in ["BUSNUM", "LOADID"]:
                                                    val = {"LOADID": load_id, "BUSNUM": int(b)}[v]
                                                    results = self.add_result(results, q, val, f"{load_id}_{b}")
                                                elif v == "BUSNAME":
                                                    ierr, val = self.psse.notona(int(b))

                                                    results = self.add_result(results, q, val, f"{load_id}_{b}")
                                            elif func_name == "loddt2":
                                                ierr, val = getattr(self.psse, func_name)(int(b), load_id, v, "ACT")

                                                results = self.add_result(results, q, val, f"{load_id}_{b}")
                                            elif func_name == "lodint":
                                                ierr, val = getattr(self.psse, func_name)(int(b), load_id, v)

                                                results = self.add_result(results, q, val, f"{load_id}_{b}")

                                            ierr, load_id = self.psse.nxtlod(int(b))

                                    elif func_name in ["macdat", "macdt2", "macnofunc", "macint"]:
                                        ierr = self.psse.inimac(int(b))
                                        ierr, mach_id = self.psse.nxtmac(int(b))
                                        while mach_id is not None:
                                            if func_name == "macnofunc":
                                                if v in ["BUSNUM", "MACID"]:
                                                    val = {"BUSNUM": int(b), "MACID": mach_id}[v]
                                                    results = self.add_result(results, q, val, f"{mach_id}_{b}")
                                                elif v == "BUSNAME":
                                                    ierr, val = self.psse.notona(int(b))
                                                    results = self.add_result(results, q, val, f"{mach_id}_{b}")
                                                elif v == "SUBNUMBER":
                                                    ierr, val = self.psse.busint(int(b), "STATION")

                                                    results = self.add_result(results, q, val, f"{mach_id}_{b}")

                                                elif v == "AREANUMBER":
                                                    ierr, val = self.psse.busint(int(b), "AREA")

                                                    results = self.add_result(results, q, val, f"{mach_id}_{b}")

                                                elif v in ["SUBLATITUDE", "SUBLONGITUDE"]:
                                                    sub_dict = {"SUBLATITUDE": "LATI", "SUBLONGITUDE": "LONG"}
                                                    ierr, val = self.psse.busint(int(b), "STATION")

                                                    if val:
                                                        ierr, val = self.psse.stadat(val, sub_dict[v])

                                                    results = self.add_result(results, q, val, b)
                                            else:
                                                ierr, val = getattr(self.psse, func_name)(int(b), mach_id, v)

                                                results = self.add_result(results, q, val, f"{mach_id}_{b}")
                                            ierr, mach_id = self.psse.nxtmac(int(b))

                                    elif func_name in ["fxsdt2", "fxsnofunc"]:
                                        ierr = self.psse.inifxs(int(b))

                                        ierr, fx_id = self.psse.nxtfxs(int(b))

                                        while fx_id is not None:
                                            if func_name == "fxsnofunc":
                                                if v in ["BUSNUM", "FXSHID"]:
                                                    val = {"BUSNUM": int(b), "FXSHID": fx_id}[v]
                                                    results = self.add_result(results, q, val, f"{fx_id}_{b}")
                                                elif v == "BUSNAME":
                                                    ierr, val = self.psse.notona(int(b))

                                                    results = self.add_result(results, q, val, f"{fx_id}_{b}")
                                            else:
                                                ierr, val = getattr(self.psse, func_name)(int(b), fx_id, v)

                                                results = self.add_result(results, q, val, f"{fx_id}_{b}")
                                            ierr, fx_id = self.psse.nxtfxs(int(b))

                                    elif func_name in ["swsdt1", "swsnofunc"]:
                                        if func_name == "swsnofunc":
                                            ierr, val = self.psse.swsint(int(b), "STATUS")

                                        else:
                                            ierr, val = getattr(self.psse, func_name)(int(b), v)

                                        if ierr == 0:
                                            if func_name == "nofunc":
                                                if v == "BUSNUM":
                                                    val = int(b)
                                                elif v == "BUSNAME":
                                                    ierr, val = self.psse.notona(int(b))

                                            results = self.add_result(results, q, val, b)

                                    elif func_name in ["brndat", "brndt2", "brnmsc", "brnint", "brnnofunc"]:
                                        ierr = self.psse.inibrx(int(b), 1)

                                        ierr, b1, ickt = self.psse.nxtbrn(int(b))

                                        ickt_string = str(ickt)
                                        while ierr == 0:
                                            if func_name == "brnnofunc":
                                                if v in ["FROMBUSNUM", "TOBUSNUM", "CIRCUIT"]:
                                                    val = {
                                                        "FROMBUSNUM": int(b),
                                                        "TOBUSNUM": int(b1),
                                                        "CIRCUIT": ickt_string,
                                                    }[v]
                                                    results = self.add_result(
                                                        results, q, val, f"{b!s}_{b1!s}_{ickt_string}"
                                                    )
                                                elif v == "FROMBUSNAME":
                                                    ierr, val = self.psse.notona(int(b))

                                                    results = self.add_result(
                                                        results, q, val, f"{b!s}_{b1!s}_{ickt_string}"
                                                    )
                                                elif v == "TOBUSNAME":
                                                    ierr, val = self.psse.notona(int(b1))

                                                    results = self.add_result(
                                                        results, q, val, f"{b!s}_{b1!s}_{ickt_string}"
                                                    )
                                                elif v in ["SUBNUMBERTO", "SUBNUMBERFROM"]:
                                                    sub_dict = {"SUBNUMBERFROM": int(b), "SUBNUMBERTO": int(b1)}
                                                    ierr, val = self.psse.busint(sub_dict[v], "STATION")

                                                    results = self.add_result(
                                                        results, q, val, f"{b!s}_{b1!s}_{ickt_string}"
                                                    )
                                                elif v == "NOMKVFROM":
                                                    ierr, val = self.psse.busdat(int(b), "BASE")

                                                    results = self.add_result(
                                                        results, q, val, f"{b!s}_{b1!s}_{ickt_string}"
                                                    )
                                                elif v == "NOMKVTO":
                                                    ierr, val = self.psse.busdat(int(b1), "BASE")

                                                    results = self.add_result(
                                                        results, q, val, f"{b!s}_{b1!s}_{ickt_string}"
                                                    )
                                                elif v == "BY":
                                                    ierr, vali = self.psse.brndt2(int(b), int(b1), "ISHNT", v)

                                                    ierr, valj = self.psse.brndt2(int(b), int(b1), "JSHNT", v)

                                                    if isinstance(vali, complex) and isinstance(valj, complex):
                                                        val = 0.5 * vali.imag + 0.5 * valj.imag
                                                    else:
                                                        val = None
                                                    results = self.add_result(
                                                        results, q, val, f"{b!s}_{b1!s}_{ickt_string}"
                                                    )

                                            else:
                                                ierr, val = getattr(self.psse, func_name)(int(b), int(b1), str(ickt), v)

                                                if irr == 0:
                                                    results = self.add_result(
                                                        results, q, val, f"{b!s}_{b1!s}_{ickt_string}"
                                                    )
                                            ierr, b1, ickt = self.psse.nxtbrn(int(b))

                                    elif func_name in ["xfrdat", "tr3dt2", "trfnofunc"]:
                                        ierr = self.psse.inibrx(int(b), 1)

                                        ierr, b1, ickt = self.psse.nxtbrn(int(b))

                                        ickt_string = str(ickt)
                                        while ierr == 0:
                                            if func_name == "xfrdat":
                                                ierr, val = getattr(self.psse, func_name)(int(b), int(b1), ickt, v)

                                                if ierr == 0:
                                                    results = self.add_result(
                                                        results, q, val, f"{b!s}_{b1!s}_{ickt_string}"
                                                    )
                                            elif func_name == "trfnofunc":
                                                if v.split("_")[1] == "2WDG":
                                                    if v in ["FROMBUSNUM_2WDG", "TOBUSNUM_2WDG", "CIRCUIT_2WDG"]:
                                                        val = {
                                                            "FROMBUSNUM_2WDG": int(b),
                                                            "TOBUSNUM_2WDG": int(b1),
                                                            "CIRCUIT_2WDG": ickt,
                                                        }[v]
                                                        results = self.add_result(
                                                            results,
                                                            q,
                                                            val,
                                                            f"{b!s}_{b1!s}_{ickt_string}",
                                                        )
                                                    elif v == "FROMBUSNAME_2WDG":
                                                        ierr, val = self.psse.notona(int(b))

                                                        results = self.add_result(
                                                            results,
                                                            q,
                                                            val,
                                                            f"{b!s}_{b1!s}_{ickt_string}",
                                                        )
                                                    elif v == "TOBUSNAME_2WDG":
                                                        ierr, val = self.psse.notona(int(b1))

                                                        results = self.add_result(
                                                            results,
                                                            q,
                                                            val,
                                                            f"{b!s}_{b1!s}_{ickt_string}",
                                                        )

                                            ierr, b1, ickt = self.psse.nxtbrn(int(b))

                                        ierr = self.psse.inibrx(int(b), 1)

                                        ierr, b1, b2, ickt = self.psse.nxtbrn3(int(b))

                                        ickt_string = str(ickt)
                                        while ierr == 0:
                                            if func_name == "tr3dt2":
                                                ierr, val = getattr(self.psse, func_name)(
                                                    int(b), int(b1), int(b2), ickt, v
                                                )

                                                if ierr == 0:
                                                    results = self.add_result(
                                                        results,
                                                        q,
                                                        val,
                                                        f"{b!s}_{b1!s}_{b2!s}_{ickt_string}",
                                                    )
                                            elif func_name == "trfnofunc":
                                                if v.split("_")[1] == "3WDG":
                                                    if v in [
                                                        "FROMBUSNUM_3WDG",
                                                        "TOBUSNUM_3WDG",
                                                        "TOBUS2NUM_3WDG",
                                                        "CIRCUIT_3WDG",
                                                    ]:
                                                        val = {
                                                            "FROMBUSNUM_3WDG": int(b),
                                                            "TOBUSNUM_3WDG": int(b1),
                                                            "CIRCUIT_3WDG": ickt,
                                                            "TOBUS2NUM_3WDG": int(b2),
                                                        }[v]
                                                        results = self.add_result(
                                                            results,
                                                            q,
                                                            val,
                                                            f"{b!s}_{b1!s}_{ickt_string}",
                                                        )
                                                    elif v == "FROMBUSNAME_3WDG":
                                                        ierr, val = self.psse.notona(int(b))

                                                        results = self.add_result(
                                                            results,
                                                            q,
                                                            val,
                                                            f"{b!s}_{b1!s}_{ickt_string}",
                                                        )
                                                    elif v == "TOBUSNAME_3WDG":
                                                        ierr, val = self.psse.notona(int(b1))

                                                        results = self.add_result(
                                                            results,
                                                            q,
                                                            val,
                                                            f"{b!s}_{b1!s}_{ickt_string}",
                                                        )
                                                    elif v == "TOBUS2NAME_3WDG":
                                                        ierr, val = self.psse.notona(int(b1))

                                                        results = self.add_result(
                                                            results,
                                                            q,
                                                            val,
                                                            f"{b!s}_{b1!s}_{ickt_string}",
                                                        )
                                            ierr, b1, b2, ickt = self.psse.nxtbrn3(int(b))

        return results

    def add_result(self, results_dict, class_name, value, label):
        if value or value == 0:
            if class_name not in results_dict:
                results_dict[class_name] = {}
            results_dict[class_name][label] = value
        else:
            if class_name not in results_dict:
                results_dict[class_name] = {}
            results_dict[class_name][label] = None
        return results_dict

    def convert_load(self, bus_subsystem=None):
        if self.settings.loads.convert:
            p1 = self.settings.loads.active_load.constant_current_percentage
            p2 = self.settings.loads.active_load.constant_admittance_percentage
            q1 = self.settings.loads.reactive_load.constant_current_percentage
            q2 = self.settings.loads.reactive_load.constant_admittance_percentage
            if bus_subsystem:
                self.psse.conl(bus_subsystem, 0, 1, [0, 0], [p1, p2, q1, q2])  # initialize for load conversion.
                self.psse.conl(bus_subsystem, 0, 2, [0, 0], [p1, p2, q1, q2])  # convert loads.
                self.psse.conl(bus_subsystem, 0, 3, [0, 0], [p1, p2, q1, q2])  # postprocessing housekeeping.
            else:
                self.psse.conl(0, 1, 1, [0, 0], [p1, p2, q1, q2])  # initialize for load conversion.
                self.psse.conl(0, 1, 2, [0, 0], [p1, p2, q1, q2])  # convert loads.
                self.psse.conl(0, 1, 3, [0, 0], [p1, p2, q1, q2])  # postprocessing housekeeping.

    def update_object(self, dtype, bus, element_id, values: dict):
        val = sum(list(values.values()))
        if val > -VALUE_UPDATE_BOUND and val < VALUE_UPDATE_BOUND:
            if dtype == WritableModelTypes.LOAD.value:
                ierr = self.psse.load_chng_5(ibus=int(bus), id=element_id, **values)
            elif dtype == WritableModelTypes.GENERATOR.value:
                ierr = self.psse.induction_machine_data(ibus=int(bus), id=element_id, **values)
            elif dtype == WritableModelTypes.MACHINE.value:
                ierr = self.psse.machine_data_2(i=int(bus), id=element_id, **values)
            elif dtype == WritableModelTypes.PLANT.value:
                ierr = self.psse.plant_data_4(ibus=int(bus), inode=element_id, **values)
            else:
                ierr = 1

            if ierr == 0:
                logger.info(f"Profile Manager: {dtype} '{element_id}' on bus '{bus}' has been updated. {values}")
            else:
                logger.error(f"Profile Manager: Error updating {dtype} '{element_id}' on bus '{bus}'.")

    def has_converged(self):
        return self.psse.solved()



"""
CLI to run the PyDSS server
"""
import click

from pypsse.api.server import run_server


@click.option(
    "-p",
    "--port",
    default=9090,
    show_default=True,
    help="Socket port for the server",
)
@click.option(
    "-h",
    "--host-ip",
    default="127.0.0.1",
    show_default=True,
    help="IP for the server",
)
@click.command()
def serve(host_ip="127.0.0.1", port=9090):
    """Run a PyPSSE RESTful API server."""
    run_server(host_ip, port)



# Standard imports
import os

from loguru import logger

# Internal imports
from pypsse.modes.abstract_mode import AbstractMode


class Static(AbstractMode):
    def __init__(self, psse, dyntools, settings, export_settings, subsystem_buses, raw_data):
        "Class defination for steady-state simulation mode"

        super().__init__(psse, dyntools, settings, export_settings, subsystem_buses, raw_data)
        self.time = settings.simulation.start_time
        self._StartTime = settings.simulation.start_time
        self.incTime = settings.simulation.simulation_step_resolution

    def init(self, bussubsystems):
        super().init(bussubsystems)
        self.initialization_complete = True

    def step(self, _):
        "Increments the simulation"
        ierr = self.psse.fnsl()
        # check if powerflow completed successfully
        if ierr == 0:
            self.time = self.time + self.incTime
        else:
            msg = f"Error code {ierr} returned from PSSE while running powerflow, please follow \
                            PSSE doumentation to know more about error"
            raise Exception(msg)

    def resolve_step(self):
        "Resolves the current time step"
        ierr = self.psse.fnsl()
        if ierr > 0:
            msg = f"Error code {ierr} returned from PSSE while running powerflow, please follow \
                                        PSSE doumentation to know more about error"
            raise Exception(msg)

    def get_time(self):
        "Returns current simulator time"
        return self.time

    def get_total_seconds(self):
        "Returns total simulation time"
        return (self.time - self._StartTime).total_seconds()

    def get_step_size_cec(self):
        "Returns simulation timestep resolution"
        return self.settings.simulation.simulation_step_resolution.total_seconds()

    def export(self):
        "Exports simulation results"
        logger.debug("Starting export process. Can take a few minutes for large files")
        excelpath = os.path.join(self.export_path, self.settings["Excel file"])
        achnf = self.dyntools.CHNF(self.outx_path)
        achnf.xlsout(channels="", show=False, xlsfile=excelpath, outfile="", sheet="Sheet1", overwritesheet=True)
        logger.debug(f"{self.settings.export.excel_file} exported")



import json
import os
from distutils.dir_util import copy_tree
from pathlib import Path
from shutil import copy

import pandas as pd
import toml
from loguru import logger

from pypsse.common import (
    CASESTUDY_FOLDER,
    DEFAULT_COORDINATES_FILE,
    DEFAULT_EXCEL_FILE,
    DEFAULT_GRAPH_FILE,
    DEFAULT_LOG_FILE,
    DEFAULT_OUT_FILE,
    DEFAULT_OUTX_FILE,
    DEFAULT_PROFILE_MAPPING_FILENAME,
    DEFAULT_SUBSCRIPTION_FILENAME,
    DEFAULTS_FOLDER,
    EXPORTS_SETTINGS_FILENAME,
    PROFILES_FOLDER,
    SIMULATION_SETTINGS_FILENAME,
)
from pypsse.enumerations import SubscriptionFileRequiredColumns
from pypsse.models import ExportFileOptions, ProjectDefination, SimulationSettings
from pypsse.profile_manager.profile_store import ProfileManager

class Project:
    "This class defines the structure of a PyPSSE project"

    def __init__(self):
        self.basepath = Path(__file__).parent

    def create(
        self,
        parent_path: Path,
        project_name: str,
        psse_folder: Path,
        simulation_settings_file: Path,
        export_settings_file: Path,
        profile_store_file: Path,
        profile_mapping_file: Path,
        overwrite: bool = True,
        autofill: bool = True,
    ):
        """The methods creates a new PyPSSE project

        Args:
            parent_path (Path): path to new pypsse project
            project_name (str): project name
            psse_folder (Path): _description_
            simulation_settings_file (Path): simulation settings toml file path
            export_settings_file (Path): export settings toml file path
            profile_store_file (Path): path to a valid Profiles.hdf5 file (Contains profiles for time series simulations)
            profile_mapping_file (Path): path to a valid Profile_mapping.toml file (used to map profile to PSSE elements)
            overwrite (bool, optional): Attempt to auto fill settings. (Verify manually settings file is correct). Defaults to True.
            autofill (bool, optional): Overwrite project is it already exists. Defaults to True.
        """

        self.project_path = Path(parent_path) / project_name

        exports_dict = toml.load(self.basepath / DEFAULTS_FOLDER / EXPORTS_SETTINGS_FILENAME)
        export_settings = ExportFileOptions(**exports_dict)

        if export_settings_file:
            export_settings_file = Path(export_settings_file)
            assert export_settings_file.exists(), f"Export file '{export_settings_file}' does not exist"
            new_export_settings = toml.load(export_settings_file)
            export_settings.update(**new_export_settings)

        sim_setting_dict = toml.load(self.basepath / DEFAULTS_FOLDER / SIMULATION_SETTINGS_FILENAME)
        sim_setting_dict["simulation"]["project_path"] = str(self.project_path)
        simulation_settings = SimulationSettings(**sim_setting_dict)

        if simulation_settings_file:
            simulation_settings_file = Path(simulation_settings_file)
            assert simulation_settings_file.exists(), f"Export file '{simulation_settings_file}' does not exist"
            sim_setting_dict = toml.load(simulation_settings_file)
            simulation_settings.update(**sim_setting_dict)

        simulation_settings.simulation.project_path = Path(parent_path) / project_name

        self.project = ProjectDefination(
            project_name=project_name,
            simulation_settings=simulation_settings,
            export_settings=export_settings,
            overwrite=overwrite,
            autofill=autofill,
        )

        self._create_folders()

        if psse_folder:
            psse_files = self._copy_psse_project_files(psse_folder)
            if autofill and psse_files:
                self._autofill_settings(psse_files, profile_store_file, profile_mapping_file)

        self._update_export_files()
        self._write_setting_files()

    def _update_export_files(self):
        """sets up export file paths"""
        self.project.simulation_settings.export.out_file = DEFAULT_OUT_FILE
        self.project.simulation_settings.export.outx_file = DEFAULT_OUTX_FILE
        self.project.simulation_settings.export.log_file = DEFAULT_LOG_FILE
        self.project.simulation_settings.export.excel_file = DEFAULT_EXCEL_FILE
        self.project.simulation_settings.export.coordinate_file = DEFAULT_COORDINATES_FILE
        self.project.simulation_settings.export.networkx_graph_file = DEFAULT_GRAPH_FILE

    def _psse_project_file_dict(self, path: Path) -> dict:
        """Creates a mapping of all project files in a folder

        Args:
            path (Path): path (folder) to existing psse project

        Returns:
            dict: mapping of file types to paths
        """

        file_dict = {}
        for _, _, files in os.walk(path):
            for file in files:
                _, ext = file.split(".")
                if ext not in file_dict:
                    file_dict[ext.lower()] = [file]
                else:
                    file_dict[ext.lower()].append(file)
        return file_dict

    def _copy_psse_project_files(self, psse_folder: Path) -> dict:
        """Copies psse files to a new project

        Args:
            psse_folder (Path): path (folder) to existing psse project

        Raises:
            FileExistsError: raised if provided project path does not exist

        Returns:
            dict: mapping of file types to paths
        """

        psse_folder = Path(psse_folder)
        if psse_folder.exists():
            new_path = self.project_path / CASESTUDY_FOLDER
            copy_tree(str(psse_folder.absolute()), str(new_path.absolute()))
            psse_files = self._psse_project_file_dict(new_path)
        else:
            msg = f"PSSE project path does not exist. ({psse_folder}) {os.getcwd()}"
            raise FileExistsError(msg)
        return psse_files

    def _write_setting_files(self):
        """serialized simulation and export setting  files for the new project"""
        sim_file_path = self.project_path / SIMULATION_SETTINGS_FILENAME
        with open(sim_file_path, "w") as f:
            toml.dump(json.loads(self.project.simulation_settings.model_dump_json()), f)
            logger.info(f"writing file : {sim_file_path!s}")

        export_file_path = self.project_path / EXPORTS_SETTINGS_FILENAME
        with open(export_file_path, "w") as f:
            toml.dump(json.loads(self.project.export_settings.model_dump_json()), f)
            logger.info(f"writing file : {export_file_path!s}")

    def _create_folders(self):
        """Creates folder structure for a new project. Older project can be over-written

        Raises:
            FileExistsError: raised if provided folder path does not exist
        """

        for folder in self.project.project_folders:
            project_folder = self.project_path / folder.value
            if project_folder.exists() and not self.project.overwrite:
                msg = "Project folder already exists. Set overwrite=true to overwrite existing projects"
                raise FileExistsError(msg)
            elif not project_folder.exists():
                project_folder.mkdir(parents=True, exist_ok=True)
                logger.info(f"folder created: {project_folder!s}")

    def _autofill_settings(self, psse_files: dict, profile_store_file: Path, profile_mapping_file: Path):
        """The method auto populates fields for a new PyPSSE project

        Args:
            psse_files (dict): mapping of file type to file path
            profile_store_file (Path): path to profile store file (hdf5)
            profile_mapping_file (Path): path to profile mapping file (toml)
        """

        self._update_setting("sav", "case_study", psse_files)
        self._update_setting("raw", "raw_file", psse_files)
        self._update_setting("snp", "snp_file", psse_files)
        self._update_setting("dyr", "dyr_file", psse_files)
        self._update_setting("gic", "gic_file", psse_files)
        self._update_setting("rwm", "rwm_file", psse_files)

        if "dll" in psse_files:
            self.project.simulation_settings.simulation.user_models = psse_files["dll"]
            logger.info(f"user_models={psse_files['dll']}")
        else:
            logger.info("No DLL files found in project path")

        if "idv" in psse_files:
            self.project.simulation_settings.simulation.setup_files = psse_files["idv"]
            logger.info(
                f"setup_files={psse_files['idv']}"
                f"\nSequence of IDV setup files is important. Manually change in TOML file if needed"
            )
        else:
            logger.info("No IDV files found in project path")

        store_path = self.project_path / PROFILES_FOLDER
        if profile_store_file and Path(profile_store_file).exists():
            assert Path(profile_store_file).suffix.lower() == ".hdf5", "Store file should be a valid hdf5 file"
            copy(profile_store_file, store_path)
        else:
            ProfileManager(None, self.project.simulation_settings)

        if profile_mapping_file and Path(profile_mapping_file).exists():
            assert Path(profile_mapping_file).suffix.lower() == ".toml", "Profile mapping should be a valid toml file"
            copy(profile_mapping_file, store_path)
            self.project.simulation_settings.simulation.use_profile_manager = True
        else:
            # TODO: auto generate mapping file from bus subsystem files
            with open(os.path.join(store_path, DEFAULT_PROFILE_MAPPING_FILENAME), "w") as _:
                pass
            self.project.simulation_settings.simulation.use_profile_manager = False

        self._create_default_sub_file()

    def _create_default_sub_file(self):
        """Method creates a subscription file for the HELICS interface"""

        subscription_fields = [x.value for x in SubscriptionFileRequiredColumns]
        data = pd.DataFrame({}, columns=list(subscription_fields))
        data.to_csv(self.project_path / DEFAULT_SUBSCRIPTION_FILENAME, index=False)
        logger.info("Creating subscription template")
        self.project.simulation_settings.simulation.subscriptions_file = DEFAULT_SUBSCRIPTION_FILENAME

    def _update_setting(self, f_type: str, key: str, psse_files: dict):
        """updates settings for the new project

        Args:
            f_type (str): file type
            key (str): simulation setting to update
            psse_files (dict): mapping of file type to file path
        """
        if f_type in psse_files:
            relevent_files = psse_files[f_type]
            setattr(self.project.simulation_settings.simulation, key, relevent_files[0])
            logger.info(f"Settings:{key}={relevent_files[0]}")
            if len(relevent_files) > 1:
                logger.warning(
                    f"More than one file with extension {f_type} exist."
                    f"\nFiles found: {relevent_files}"
                    f"\nManually update the settings toml file"
                )
        else:
            logger.warning(f"No file with extension '{f_type}'")



import numpy as np
from loguru import logger

from pypsse.models import ExportSettings, SimulationSettings
from pypsse.modes.abstract_mode import AbstractMode
from pypsse.modes.constants import DYNAMIC_ONLY_PPTY, converter, dyn_only_options
from pypsse.utils.dynamic_utils import DynamicUtils


class Snap(AbstractMode, DynamicUtils):
    "Class defination for snapshat simulation mode (uses snp and sav files)"

    def __init__(
        self,
        psse,
        dyntools,
        settings: SimulationSettings,
        export_settings: ExportSettings,
        subsystem_buses,
        raw_data,
    ):
        super().__init__(psse, dyntools, settings, export_settings, subsystem_buses, raw_data)
        self.time = settings.simulation.start_time
        self._StartTime = settings.simulation.start_time
        self.incTime = settings.simulation.simulation_step_resolution
        self.init(subsystem_buses)

    def init(self, bus_subsystems):
        "Initializes the simulation"
        super().init(bus_subsystems)

        self.iter_const = 100.0
        self.xTime = 0

        ierr = self.psse.case(str(self.settings.simulation.case_study))

        self.load_setup_files()
        self.convert_load()

        logger.info(f"Load snap file: {self.settings.simulation.snp_file}")
        ierr = self.psse.rstr(str(self.settings.simulation.snp_file))
        #

        # The following logic only runs when the helics interface is enabled
        self.disable_load_models_for_coupled_buses()
        self.disable_generation_for_coupled_buses()
        # self.save_model()
        ############# ------------------------------------- ###############

        ierr = self.psse.strt_2([0, 1], str(self.settings.export.outx_file))

        if ierr == 1:
            self.psse.cong(0)
            ierr = self.psse.strt_2([0, 1], str(self.settings.export.outx_file))

        elif ierr > 1:
            msg = "Error starting simulation"
            raise Exception(msg)

        self.load_user_defined_models()

        if self.settings.helics and self.settings.helics.cosimulation_mode:
            if self.settings.helics.iterative_mode:
                sim_step = self.settings.simulation.psse_solver_timestep.total_seconds() / self.iter_const
            else:
                sim_step = self.settings.simulation.psse_solver_timestep.total_seconds()
        else:
            sim_step = self.settings.simulation.psse_solver_timestep.total_seconds()

        self.psse.dynamics_solution_param_2(
            [60, self._i, self._i, self._i, self._i, self._i, self._i, self._i],
            [0.4, self._f, sim_step, self._f, self._f, self._f, self._f, self._f],
        )

        self.psse.delete_all_plot_channels()

        self.setup_all_channels()

        logger.debug("pyPSSE initialization complete!")
        self.initialization_complete = True
        return self.initialization_complete

    def step(self, t):
        "Increments the simulation"
        self.time = self.time + self.incTime
        self.xTime = 0
        return self.psse.run(0, t, 1, 1, 1)

    def resolve_step(self, t):
        "Resolves the current time step"
        self.xTime += 1
        return self.psse.run(0, t + self.xTime * self.incTime / self.iter_const, 1, 1, 1)

    def get_load_indices(self, bus_subsystems):
        "Returns load indices"
        all_bus_ids = {}
        for bus_subsystem_id in bus_subsystems.keys():
            load_info = {}
            ierr, load_data = self.psse.aloadchar(bus_subsystem_id, 1, ["ID", "NAME", "EXNAME"])

            load_data = np.array(load_data)
            ierr, bus_data = self.psse.aloadint(bus_subsystem_id, 1, ["NUMBER"])

            bus_data = bus_data[0]
            for i, bus_id in enumerate(bus_data):
                load_info[bus_id] = {
                    "Load ID": load_data[0, i],
                    "Bus name": load_data[1, i],
                    "Bus name (ext)": load_data[2, i],
                }
            all_bus_ids[bus_subsystem_id] = load_info
        return all_bus_ids

    def get_time(self):
        "Returns current simulator time"
        return self.time

    def get_total_seconds(self):
        "Returns total simulation time"
        return (self.time - self._StartTime).total_seconds()

    def get_step_size_cec(self):
        "Returns simulation timestep resolution"
        return self.settings.simulation.simulation_step_resolution.total_seconds()

    @converter
    def read_subsystems(self, quantities, subsystem_buses, ext_string2_info=None, mapping_dict=None):
        "Queries the result container for current results"
        if ext_string2_info is None:
            ext_string2_info = {}
        if mapping_dict is None:
            mapping_dict = {}
        results = super().read_subsystems(
            quantities, subsystem_buses, mapping_dict=mapping_dict, ext_string2_info=ext_string2_info
        )

        poll_results = self.poll_channels()
        results.update(poll_results)
        """ Add """
        for class_name, var_list in quantities.items():
            if class_name in dyn_only_options:
                for v in var_list:
                    if v in DYNAMIC_ONLY_PPTY[class_name]:
                        for func_name in dyn_only_options[class_name]:
                            if v in dyn_only_options[class_name][func_name]:
                                con_ind = dyn_only_options[class_name][func_name][v]
                                for bus in subsystem_buses:
                                    if class_name == "Loads":
                                        ierr = self.psse.inilod(int(bus))

                                        ierr, ld_id = self.psse.nxtlod(int(bus))

                                        if ld_id is not None:
                                            ierr, con_index = getattr(self.psse, func_name)(
                                                int(bus), ld_id, "CHARAC", "CON"
                                            )

                                            if con_index is not None:
                                                act_con_index = con_index + con_ind
                                                ierr, value = self.psse.dsrval("CON", act_con_index)

                                                res_base = f"{class_name}_{v}"
                                                if res_base not in results:
                                                    results[res_base] = {}
                                                obj_name = f"{bus}_{ld_id}"
                                                results[res_base][obj_name] = value
            else:
                logger.warning("Extend function 'read_subsystems' in the Snap class (Snap.py)")

        return results



"This module manages contingency modeling in PyPSSE"


from abc import ABCMeta
from typing import List, Union

from loguru import logger

# from pyPSSE import
from pypsse.models import (
    BusFault,
    Contingencies,
    BusTrip,
    LineFault,
    LineTrip,
    MachineTrip,
    SimulationSettings,
)


def add_contingency(contingency, cont_dict, dt, system_contingencies):
    "Adds a new contingency"
    ontingency_type = contingency.__class__.__name__
    if ontingency_type in contingencies:
        system_contingencies.append(contingencies[ontingency_type](**cont_dict))
    else:
        msg = "Invalid contingency type. Valid values are: {}".format(
            ",".join(contingencies.keys())
        )
        raise Exception(msg)
    return system_contingencies, dt


class BaseFault:
    "Base class defination for all fault types"
    __metaclass__ = ABCMeta

    requirements = []
    fault_settings = {}
    fault_method = ""
    element = None

    def __init__(self, psse, settings, contingency_type):
        self.contingency_type = contingency_type
        self.settings = settings
        self.psse = psse
        self.enabled = False
        self.tripped = False

    def update(self, t: float):
        """updates a fault event

        Args:
            t (float): simuation time in seconds
        """
        self.t = t
        if hasattr(self.settings, "duration"):
            if (
                self.settings.time + self.settings.duration
                > t
                >= self.settings.time
                and not self.enabled
            ):
                self.enabled = True
                self.enable_fault()
            if (
                t >= self.settings.time + self.settings.duration
                and self.enabled
            ):
                self.enabled = False
                self.disable_fault()
        elif (
            not hasattr(self.settings, "duration")
            and t >= self.settings.time
            and not self.tripped
        ):
            self.enable_fault()
            self.tripped = True

    def enable_fault(self):
        """enables a fault event"""
        err = getattr(self.psse, self.fault_method)(**self.fault_settings)
        if err:
            logger.warning(
                f"Unable to enable {self.fault_method} at element {self.element}"
            )
        else:
            logger.debug(
                f"{self.fault_method} applied to {self.element} at time {self.t} seconds"
            )

    def disable_fault(self):
        """disables a fault event"""
        err = self.psse.dist_clear_fault()
        if err:
            logger.warning(
                f"Unable to clear {self.fault_method} at element {self.element}"
            )
        else:
            logger.debug(
                f"{self.fault_method} cleared at element {self.element} at time {self.t} seconds"
            )

    def is_enabled(self):
        """Returns enabled status

        Returns:
            _type_: rue if the fault object is enabled else false
        """
        return self.enabled

    def is_tripped(self) -> bool:
        """Returns trip status

        Returns:
            bool: true if the fault object is tripped else false
        """
        return self.tripped


class BusFaultObject(BaseFault):
    "Class defination for a bus fault"
    fault_method = "dist_bus_fault"
    fault_settings = {}

    def __init__(self, psse: object, settings: BusFault, contingency_type: str):
        """bus fault object

        Args:
            psse (object): simulator type
            settings (BusFault): bus fault object
            contingency_type (str): contingency type
        """
        super().__init__(psse, settings, contingency_type)
        self.fault_settings["ibus"] = settings.bus_id
        self.fault_settings["units"] = 3
        self.fault_settings["values"] = settings.fault_impedance
        self.fault_settings["basekv"] = 0.0
        self.element = settings.bus_id


class LineFaultObject(BaseFault):
    "Class defination for a line fault"
    fault_method = "dist_branch_fault"
    fault_settings = {}

    def __init__(
        self, psse: object, settings: LineFault, contingency_type: str
    ):
        """line fault model

        Args:
            psse (object): simulator instance
            settings (LineFault): line fault model
            contingency_type (str): contingecy type
        """
        super().__init__(psse, settings, contingency_type)
        self.fault_settings["ibus"] = settings.bus_ids[0]
        self.fault_settings["jbus"] = settings.bus_ids[1]
        self.fault_settings["id"] = settings.bus_ids[2]
        self.fault_settings["units"] = 3
        self.fault_settings["values"] = settings.fault_impedance
        self.fault_settings["basekv"] = 0.0
        self.element = settings.bus_ids


class LineTripObject(BaseFault):
    "Class defination for a line trip"
    fault_method = "dist_branch_trip"
    fault_settings = {}

    def __init__(self, psse: object, settings: LineTrip, contingency_type: str):
        """line trip model

        Args:
            psse (object): simulator instance
            settings (LineTrip): line trip model
            contingency_type (str): contingency type
        """
        super().__init__(psse, settings, contingency_type)
        self.fault_settings["ibus"] = settings.bus_ids[0]
        self.fault_settings["jbus"] = settings.bus_ids[1]
        self.element = settings.bus_ids


class BusTripObject(BaseFault):
    "Class defination for a bus trip"
    fault_method = "dist_bus_trip"
    fault_settings = {}

    def __init__(self, psse: object, settings: BusTrip, contingency_type: str):
        """Bus trip contingency

        Args:
            psse (object): simulator instance
            settings (BusTrip): bus trip model
            contingency_type (str): type of contingency
        """
        super().__init__(psse, settings, contingency_type)
        self.fault_settings["ibus"] = settings.bus_id
        self.element = settings.bus_id


class MachineTripObject(BaseFault):
    "Class defination for a machine fault"
    fault_method = "dist_machine_trip"
    fault_settings = {}

    def __init__(
        self, psse: object, settings: MachineTrip, contingency_type: str
    ):
        """Machine trip contingency

        Args:
            psse (object): simulator instance
            settings (MachineTrip): machine trip model
            contingency_type (str): type of contingency
        """
        super().__init__(psse, settings, contingency_type)
        self.fault_settings["ibus"] = settings.bus_id
        self.fault_settings["id"] = settings.machine_id
        self.element = settings.bus_id


contingencies = {
    "BusFault": BusFaultObject,
    "LineFault": LineFaultObject,
    "LineTrip": LineTripObject,
    "BusTrip": BusTripObject,
    "MachineTrip": MachineTripObject,
}


def build_contingencies(
    psse: object, contingencies_: Union[Contingencies, SimulationSettings]
) -> List[
    Union[
        BusFaultObject,
        BusTripObject,
        LineFaultObject,
        LineTripObject,
        MachineTripObject,
    ]
]:
    """Builds all contingencies defined in the settings file

    Args:
        psse (object): simulator instance
        settings (SimulationSettings): simulation settings

    Returns:
        List[Union[BusFaultObject, BusTripObject, LineFaultObject, LineTripObject, MachineTripObject]]: list of contingencies
    """

    system_contingencies = []
    if contingencies_.contingencies:
        for contingency in contingencies_.contingencies:
            contingency_type = contingency.__class__.__name__
            if contingency_type in contingencies:
                system_contingencies.append(
                    contingencies[contingency_type](
                        psse, contingency, contingency_type
                    )
                )
                logger.debug(f'Contingency of type "{contingency_type}" added')
            else:
                logger.warning(
                    "Invalid contingency type. Valid values are: {}".format(
                        ",".join(contingencies)
                    )
                )
    else:
        logger.debug("No contingencies to build")
    return system_contingencies



"""CLI to create a new PyPSSE project"""

import os

import click
import toml

from pypsse.project import Project


@click.argument(
    "path",
)
@click.option(
    "-p",
    "--project",
    required=True,
    help="project name",
)
@click.option(
    "-F",
    "--psse-project-folder",
    default=None,
    required=False,
    type=click.Path(exists=True),
    help="PSS/E project folder path",
)
@click.option(
    "-f",
    "--simulation-file",
    required=False,
    show_default=True,
    default="",
    help="Simulation settings toml file path",
)
@click.option(
    "-e",
    "--export-settings-file",
    default="",
    help="Export settings toml file path",
)
@click.option(
    "-s",
    "--profile-store",
    default="",
    help="Path to a valid Profiles.hdf5 file (Contains profiles for time series simulations)",
)
@click.option(
    "-m",
    "--profile-mapping",
    default="",
    help="Path to a valid Profile_mapping.toml file (used to map profile to PSSE elements)",
)
@click.option(
    "-a",
    "--autofill",
    help="Attempt to auto fill settings. (Verify manually settings file is correct)",
    is_flag=True,
    default=True,
    show_default=True,
)
@click.option(
    "-o",
    "--overwrite",
    help="Overwrite project is it already exists",
    is_flag=True,
    default=True,
    show_default=True,
)
@click.command()
def create_project(
    path=None,
    project=None,
    psse_project_folder=None,
    simulation_file=None,
    export_settings_file=None,
    profile_store=None,
    profile_mapping=None,
    overwrite=None,
    autofill=None,
):
    """Create a new PyPSSE project."""
    if os.path.exists(path):
        s_settings = toml.load(simulation_file) if simulation_file else {}
        e_settings = toml.load(export_settings_file) if export_settings_file else {}
        # TODO: Validate settings
        a = Project()
        a.create(
            path,
            project,
            psse_project_folder,
            s_settings,
            e_settings,
            profile_store,
            profile_mapping,
            overwrite,
            autofill,
        )



import copy
import datetime

import numpy as np
from loguru import logger

from pypsse.profile_manager.common import PROFILE_VALIDATION


class Profile:
    "Class defination fora single profile"

    DEFAULT_SETTINGS = {"multiplier": 1, "normalize": False, "interpolate": False}

    def __init__(self, profile_obj, solver, mapping_dict, buffer_size=10, neglect_year=True):
        self.value_settings = {f"{x['bus']}__{x['id']}": {**self.DEFAULT_SETTINGS, **x} for x in mapping_dict}
        self.mapping_dict = mapping_dict
        self.buffer_size = buffer_size
        self.buffer = np.zeros(buffer_size)
        self.profile = profile_obj
        self.neglect_year = neglect_year
        self.solver = solver
        self.attrs = self.profile.attrs
        s = self.attrs["sTime"].decode()
        stime = s if "." in s else s + ".00"
        e = self.attrs["eTime"].decode()
        etime = e if "." in e else e + ".00"
        self.stime = datetime.datetime.strptime(stime, "%Y-%m-%d %H:%M:%S.%f").astimezone(None)
        self.etime = datetime.datetime.strptime(etime, "%Y-%m-%d %H:%M:%S.%f").astimezone(None)
        self.sim_res = self.solver.get_step_size_cec()
        self.time = copy.deepcopy(self.solver.get_time())
        self.columns = self.attrs["units"]
        self.dtype = self.attrs["type"].decode()

    def update(self, update_object_properties=True):
        "Returns value at the current timestep in the given profile"
        self.time = copy.deepcopy(self.solver.get_time()).astimezone(None)
        if self.time < self.stime or self.time > self.etime:
            value = np.array([0] * len(self.profile[0]))
            value1 = np.array([0] * len(self.profile[0]))
        else:
            dt = (self.time - self.stime).total_seconds()
            n = int(dt / self.attrs["resTime"])
            value = np.array(list(self.profile[n]))
            try:
                valuen1 = np.array(list(self.profile[n + 1]))
            except Exception as _:
                valuen1 = value

            dt2 = (
                self.time - (self.stime + datetime.timedelta(seconds=int(n * self.attrs["resTime"])))
            ).total_seconds()
            value1 = value + (valuen1 - value) * dt2 / self.attrs["resTime"]

        if update_object_properties:
            for obj_name in self.value_settings:
                bus, object_id = obj_name.split("__")
                if self.value_settings[obj_name]["interpolate"]:
                    value = value1
                mult = self.value_settings[obj_name]["multiplier"]
                if isinstance(mult, list):
                    mult = np.array(mult)
                if self.value_settings[obj_name]["normalize"]:
                    value_f = value / self.attrs["max"] * mult
                else:
                    value_f = value * mult
                value_f = self.fill_missing_values(value_f)
                self.solver.update_object(self.dtype, bus, object_id, value_f)
        return value

    def fill_missing_values(self, value):
        "Fixes issues in profile data"
        idx = [f"realar{PROFILE_VALIDATION[self.dtype].index(c) + 1}" for c in self.columns]
        x = dict(zip(idx, list(value)))
        return x



from typing import List

from loguru import logger

from pypsse.common import MAPPED_CLASS_NAMES


class Reader:
    "Parser for indexing all PSSE model assets"

    def __init__(self, psse_instance: object):
        """creates pypsse model reader

        Args:
            psse_instance (object): simulator instance
        """
        self.psse = psse_instance
        self.buses = self.get_data("abus", tails=["int"], strings=["NUMBER"], flags=[2])
        self.loads = self.get_data("aload", tails=["int", "char"], strings=["NUMBER", "ID"], flags=[4, 4])
        self.loads = self.get_data("aload", tails=["int", "char"], strings=["NUMBER", "ID"], flags=[4, 4])
        self.fixed_stunts = self.get_data("afxshunt", tails=["int", "char"], strings=["NUMBER", "ID"], flags=[4, 4])
        self.generators = self.get_data("amach", tails=["int", "char"], strings=["NUMBER", "ID"], flags=[4, 4])
        self.branches = self.get_data(
            "abrn", tails=["int", "int", "char"], strings=["FROMNUMBER", "TONUMBER", "ID"], flags=[2, 2, 2]
        )
        self.transformers = self.get_data(
            "atr3", tails=["int", "int", "int"], strings=["WIND1NUMBER", "WIND2NUMBER", "WIND3NUMBER"], flags=[2, 2, 2]
        )
        self.area = self.get_data(
            "aarea", tails=["int", "char"], strings=["NUMBER", "AREANAME"], flags=[2, 2]
        )  # Talk to Aadil
        self.dc_branch = self.get_data(
            "a2trmdc", tails=["int", "int"], strings=["FROMNUMBER", "TONUMBER"], flags=[2, 2]
        )  # three terminal dc lines not implemented
        self.multi_term_dc = self.get_data(
            "amultitrmdc", tails=["int", "int"], strings=["VCNPOSNUMBER", "VCNNEGNUMBER"], flags=[2, 2]
        )
        self.switched_shunt = self.get_data(
            "aswsh", tails=["int", "char"], strings=["NUMBER", "DEVICENAME"], flags=[4, 4]
        )
        self.zones = self.get_data("azone", tails=["int", "char"], strings=["NUMBER", "ZONENAME"], flags=[2, 2])
        self.owners = self.get_data("aowner", tails=["int", "char"], strings=["NUMBER", "OWNERNAME"], flags=[2, 2])

    def get_data(self, func_name: str, tails: list = [], strings: list = [], flags: List[int] = []) -> list:
        """returns list of assets matching signature

        Args:
            func_name (str): _description_
            tails (list, optional): method tail. Defaults to [].
            strings (list, optional): data types. Defaults to [].
            flags (List[int], optional): list of flags for filtering. Defaults to [].

        Returns:
            list: list of asset names
        """

        array_list = []
        for tail, string, flag in zip(tails, strings, flags):
            func = getattr(self.psse, func_name.lower() + tail)
            ierr, array_1 = func(sid=-1, flag=flag, string=string)
            assert ierr == 0, f"Error code {ierr}, while running function '{func_name.lower() + tail}'"
            array_list.append([x for array in array_1 for x in array])

        logger.info(f"{func_name} count - {len(array_1)}")
        if len(array_list) == 1:
            return array_list[0]

        return list(zip(*array_list))

    def __str__(self) -> str:
        """overrides default 'print' behavior

        Returns:
            str: summary of model assets
        """
        str_name = "Model asset summary:\n"
        for model in MAPPED_CLASS_NAMES:
            str_name += f"   {model}-{len(getattr(self, model))}"
        return str_name



import logging
import os

from pypsse.common import DEFAULT_LOG_FILE
from pypsse.models import LogSettings


def get_logger(name, path, logger_options: LogSettings = None):
    "Customizes the logger for PyPSSE"
    if logger_options.clear_old_log_file:
        test = os.listdir(os.getcwd())
        for item in test:
            if item.endswith(".log"):
                try:
                    os.remove(item)
                except Exception as e:
                    logging.warning(f"Unable to delete the old log file, {e!s}")
    formatter = logging.Formatter(fmt="%(asctime)s - %(levelname)s - %(module)s - %(message)s")

    logger = logging.getLogger(name)
    logger.setLevel(logger_options.logging_level)
    if logger_options.display_on_screen:
        handler1 = logging.StreamHandler()
        handler1.setFormatter(formatter)
        logger.addHandler(handler1)
    if logger_options.log_to_external_file:
        if not os.path.exists(path):
            os.mkdir(path)
        handler2 = logging.FileHandler(filename=os.path.join(path, DEFAULT_LOG_FILE))
        handler2.setFormatter(formatter)
        logger.addHandler(handler2)
    return logger



import os

import networkx as nx
import pandas as pd
from loguru import logger

from pypsse.models import SimulationSettings


class GICParser:
    """parser for the psse GIC file"""

    valid_verions = ["3"]

    def __init__(self, settings: SimulationSettings):
        """create GIC parser object

        Args:
            settings (SimulationSettings): simulation settings
        """
        logger.debug("Starting RAW parser")

        self.settings = settings
        self.filepath = str(settings.simulation.gic_file)

        self.filehandle = open(self.filepath)
        verion = self.filehandle.readline()
        if "GICFILEVRSN=" in verion:
            verion = verion.replace("GICFILEVRSN=", "").replace("\r", "").replace("\n", "")
            if verion in self.valid_verions:
                logger.debug(f"Reading GIC file verion {verion}")
            else:
                vers = ",".join(self.valid_verions)
                logger.debug(f"Version {verion} is not supported.\nFollowing version are currently supported: {vers}")
        else:
            logger.debug("GIC file structue does not seem to be valid")

        self.get_bus_coordinates()
        self.psse_graph = nx.Graph()
        self.create_graph()
        pos = {}
        for node in self.psse_graph.nodes:
            pos[node] = [
                float(self.psse_graph.nodes[node]["latitude"]),
                float(self.psse_graph.nodes[node]["longitude"]),
            ]
        export_path = os.path.join(
            self.settings["Simulation"]["Project Path"],
            "Exports",
            self.settings["Export_settings"]["NetworkX graph file"],
        )
        nx.write_gpickle(self.psse_graph, export_path)
        # nx.draw(self.psse_graph ,pos)
        # plt.show()

    def create_graph(self):
        """creates graph representation"""

        self.parse_substation_data()
        self.parse_transformer_data()
        self.parse_branch_data()
        nx.set_node_attributes(self.psse_graph, self.bus_data)

    def parse_substation_data(self):
        """parses substation data"""

        logger.debug("Parsing substation data...")
        linedata = ""
        while True:
            linedata = self.filehandle.readline()
            if "End of Bus Substation Data" in linedata:
                break
            if self.settings["GIC_export_settings"]["include substation connections"]:
                buses = linedata.replace("\r", "").replace("\n", "")
                buses = buses.split(" ")
                if buses[0] in self.bus_data and buses[1] in self.bus_data:
                    self.psse_graph.add_edge(buses[0], buses[1])
                else:
                    logger.debug(
                        f"Error parsing substation data egde: {buses}.\nOne of the bus id does not exist in bus data"
                    )

    def parse_transformer_data(self):
        """parses transformer data"""

        logger.debug("Parsing transformer data...")
        linedata = ""
        while True:
            linedata = self.filehandle.readline()
            if "End of Transformer Data" in linedata:
                break

            if self.settings["GIC_export_settings"]["include transfomer connections"]:
                buses = linedata.replace("\r", "").replace("\n", "")
                buses = buses.split(" ")[:3]
                if buses[2] == "":
                    if buses[0] in self.bus_data and buses[1] in self.bus_data:
                        self.psse_graph.add_edge(buses[0], buses[1])
                    else:
                        logger.debug(
                            f"Error parsing transformer data egde: {buses}."
                            f"\nOne of the bus id does not exist in bus data"
                        )
                else:
                    if buses[0] in self.bus_data and buses[1] in self.bus_data:
                        self.psse_graph.add_edge(buses[0], buses[1])
                    if buses[2] in self.bus_data and buses[1] in self.bus_data:
                        self.psse_graph.add_edge(buses[1], buses[2])
                    if buses[2] in self.bus_data and buses[0] in self.bus_data:
                        self.psse_graph.add_edge(buses[2], buses[0])
                    pass

    def parse_branch_data(self):
        """parses branch data"""

        logger.debug("Parsing branch data...")
        linedata = ""
        while True:
            linedata = self.filehandle.readline()
            if "End of Branch Data" in linedata:
                break
            if self.settings["GIC_export_settings"]["include branch connections"]:
                buses = linedata.replace("\r", "").replace("\n", "")
                buses = buses.split(" ")[:2]
                if buses[0] in self.bus_data and buses[1] in self.bus_data:
                    self.psse_graph.add_edge(buses[0], buses[1])
                else:
                    logger.debug(
                        f"Error parsing branch data egde: {buses}.\nOne of the bus id does not exist in bus data"
                    )

    def get_bus_coordinates(self):
        """parses bus coordinates"""

        logger.debug("Parsing bus coordinates...")
        bus_data_headers = ["subsystem/bustype?", "latitude", "longitude", "angle?"]
        self.bus_data = {}
        linedata = ""
        start = "'"
        end = "'"
        while True:
            linedata = self.filehandle.readline()
            if "End of Substation data" in linedata:
                break

            bus_name = linedata[linedata.find(start) + len(start) : linedata.rfind(end)]
            data = linedata.replace(f" {start}{bus_name}{end}", "")
            data = data.replace("  ", " ")
            data = data.replace("  ", " ")
            data = data.split(" ")
            bus_id = data[0]

            if bus_id not in self.bus_data:
                self.bus_data[bus_id] = {}

            self.bus_data[bus_id]["bus_name"] = bus_name
            for val, label in zip(data[1:], bus_data_headers):
                self.bus_data[bus_id][label] = val

        bus_data = pd.DataFrame(self.bus_data).T
        export_path = os.path.join(
            self.settings["Simulation"]["Project Path"], "Exports", self.settings["Export_settings"]["Coordinate file"]
        )
        bus_data.to_csv(export_path)
        logger.debug(f"Bus coordinate file exported to: {export_path}")



import datetime
from pathlib import Path
from typing import Union

import h5py
import numpy as np
import pandas as pd
import toml
from loguru import logger

from pypsse.common import DEFAULT_PROFILE_MAPPING_FILENAME, DEFAULT_PROFILE_STORE_FILENAME, PROFILES_FOLDER
from pypsse.exceptions import InvalidParameterError
from pypsse.models import SimulationSettings
from pypsse.modes.dynamic import Dynamic
from pypsse.modes.pcm import ProductionCostModel
from pypsse.modes.snap import Snap
from pypsse.modes.static import Static
from pypsse.profile_manager.common import PROFILE_VALIDATION, ProfileTypes
from pypsse.profile_manager.profile import Profile


class ProfileManager:
    """Implentation for the profile manager for PyPSSE.
    Enables attacheing profilse to all PSSE objects and associated properties"""

    def __init__(
        self, solver: Union[ProductionCostModel, Snap, Static, Dynamic], settings: SimulationSettings, mode: str = "r+"
    ):
        """Creates an instance of the profile manager

        Args:
            solver (Union[ProductionCostModel, Snap, Static, Dynamic]): instance of simulation controller
            settings (SimulationSettings): simulation settings
            mode (str, optional): file update mode. Defaults to "r+".
        """

        self.solver = solver
        self.settings = settings

        file_path = settings.simulation.project_path / PROFILES_FOLDER / DEFAULT_PROFILE_STORE_FILENAME

        if file_path.exists():
            logger.info("Loading existing h5 store")
            self.store = h5py.File(file_path, mode)
        else:
            logger.info("Creating new h5 store")
            self.store = h5py.File(file_path, "w")
            for profile_group in ProfileTypes.names():
                self.store.create_group(profile_group)

    def load_data(self, file_path: Path) -> dict:
        """Load in external profile data

        Args:
            file_path (Path): path to profile mapping file

        Returns:
            dict: profile mapping dictionary
        """

        toml_dict = toml.load(file_path)
        return toml_dict

    def setup_profiles(self):
        """sets up all profiles in the profile manager

        Raises:
            Exception: raised if mapped object not found in profile DB
        """

        mapping_path = self.settings.simulation.project_path / PROFILES_FOLDER / DEFAULT_PROFILE_MAPPING_FILENAME
        if mapping_path.exists():
            self.profile_mapping = self.load_data(mapping_path)
            self.profiles = {}
            for group, profile_map in self.profile_mapping.items():
                if group in self.store:
                    grp = self.store[group]
                    for profile_name, mapping_dict in profile_map.items():
                        if profile_name in grp:
                            self.profiles[f"{group}/{profile_name}"] = Profile(
                                grp[profile_name], self.solver, mapping_dict
                            )
                        else:
                            logger.warning(rf"Group {group} \ data set {profile_name} not found in the h5 store")
                else:
                    logger.warning(f"Group {group} not found in the h5 store")
        else:
            msg = f"Profile_mapping.toml file does not exist in path {mapping_path}"
            raise Exception(msg)

    def create_dataset(
        self,
        dname: str,
        p_type: ProfileTypes,
        data: pd.DataFrame,
        start_timedate: datetime.datetime,
        resolution: float,
        _,
        info: str,
    ):
        """Create new profile datasets

        Args:
            dname (str): dateset name
            p_type (ProfileTypes): profile type
            data (pd.DataFrame): data
            start_timedate (datetime.datetime): profile start time
            resolution (float): profile resolution
            _ (_type_): _description_
            info (:str): profile description

        Raises:
            Exception: raised if dataset already exists
        """

        grp = self.store[p_type]
        if dname not in grp:
            sa, sa_type = self.df_to_sarray(data)
            dset = grp.create_dataset(
                dname, data=sa, chunks=True, compression="gzip", compression_opts=4, shuffle=True, dtype=sa_type
            )
            self.create_metadata(dset, start_time, resolution, data, list(data.columns), info, p_type)
        else:
            logger.error(f'Data set "{dname}" already exists in group "{p_type}".')
            msg = f'Data set "{dname}" already exists in group "{p_type}".'
            raise Exception(msg)

    def df_to_sarray(self, df: pd.DataFrame) -> [str, str]:
        """Enables data converson

        Args:
            df (pd.DataFrame): _description_

        Raises:
            SystemError: raised if unable to convert

        Returns:
            [str, str]: returns column name and datatype
        """

        def make_col_type(col_type, col):
            try:
                if "numpy.object_" in str(col_type.type):
                    maxlens = col.dropna().str.len()
                    if maxlens.any():
                        maxlen = maxlens.max().astype(int)
                        col_type = ("S%s" % maxlen, 1)
                    else:
                        col_type = "f2"
                return col.name, col_type
            except:
                raise SystemError("Unable to convert dataframe to np array")

        v = df.values
        types = df.dtypes
        numpy_struct_types = [make_col_type(types[col], df.loc[:, col]) for col in df.columns]
        dtype = np.dtype(numpy_struct_types)
        z = np.zeros(v.shape[0], dtype)
        for i, k in enumerate(z.dtype.names):
            # This is in case you have problems with the encoding, remove the if branch if not
            try:
                if dtype[i].str.startswith("|S"):
                    z[k] = df[k].str.encode("latin").astype("S")
                else:
                    z[k] = v[:, i]
            except:
                raise

        return z, dtype

    def add_profiles_from_csv(
        self,
        csv_file: Path,
        name: str,
        p_type: ProfileTypes,
        start_time: datetime.date,
        resolution_sec: float = 900,
        units: str = "",
        info: str = "",
    ):
        """enables profiles from existing csv files

        Args:
            csv_file (Path): path to profiles in a csv file
            name (str): profile name
            p_type (ProfileTypes): profile type
            start_time (datetime.date): profile start time
            resolution_sec (float, optional): profile resolution in seconds. Defaults to 900.
            units (str, optional): profile units. Defaults to "".
            info (str, optional): profile into. Defaults to "".

        Raises:
            ValueError: rasied if invalid profile name passed
            ValueError: rasied if invalid profile type passed
        """

        if p_type not in PROFILE_VALIDATION:
            msg = f"Valid profile types are: {list(PROFILE_VALIDATION.keys())}"
            raise ValueError(msg)
        data = pd.read_csv(csv_file)
        for c in data.columns:
            if c not in PROFILE_VALIDATION[p_type]:
                msg = f"{c} is not valid, Valid subtypes for '{p_type}' are: {PROFILE_VALIDATION[p_type]}"
                raise ValueError(msg)
        self.add_profiles(name, p_type, data, start_time, resolution_sec=resolution_sec, units=units, info=info)

    def add_profiles(
        self,
        name: str,
        data: object,
        p_type: ProfileTypes,
        start_time: datetime.date,
        resolution_sec: float = 900,
        units: str = "",
        info: str = "",
    ):
        """adds a profile to the profile manager

        Args:
            name (str): profile name
            data (object): profile data object
            p_type (ProfileTypes): profile type
            start_time (datetime.date): profile start time
            resolution_sec (float, optional): profile resolution in seconds. Defaults to 900.
            units (str, optional): profile units. Defaults to "".
            info (str, optional): profile into. Defaults to "".

        Raises:
            InvalidParameterError: raised if start_time not a datetime object
            InvalidParameterError: raised if invalid profile type passed
        """

        if type(start_time) is not datetime.datetime:
            msg = "start_time should be a python datetime object"
            raise InvalidParameterError(msg)
        if p_type not in ProfileTypes.names():
            msg = f"Valid values for p_type are {ProfileTypes.names()}"
            raise InvalidParameterError(msg)
        self.create_dataset(name, p_type, data, start_time, resolution_sec, units=units, info=info)

    def create_metadata(
        self,
        d_set: str,
        start_time: datetime.date,
        resolution: float,
        data: object,
        units: str,
        info: str,
        p_type: ProfileTypes,
    ):
        """adds a metadata to a new profile

        Args:
            d_set (str): dataset name
            start_time (datetime.date): profile start time
            resolution (float): profile resolution
            data (object): profile data object
            units (str): profile units
            info (str): profile info
            p_type (ProfileTypes): profile type
        """

        metadata = {
            "sTime": str(start_time),
            "eTime": str(start_time + datetime.timedelta(seconds=resolution * len(data))),
            "resTime": resolution,
            "npts": len(data),
            "min": data.min(),
            "max": data.max(),
            "mean": np.mean(data),
            "units": units,
            "info": info,
            "type": p_type,
        }
        for key, value in metadata.items():
            if isinstance(value, str):
                value_mod = np.string_(value)
            else:
                value_mod = value
            d_set.attrs[key] = value_mod

    def update(self) -> dict:
        """returns data for the current timestep for all mapped profiles

        Returns:
            dict: values for profiles at the current time step
        """

        results = {}
        for profile_name, profile_obj in self.profiles.items():
            result = profile_obj.update()
            results[profile_name] = result
        return results

    def __del__(self):
        self.store.flush()



"""
CLI to run a PyDSS project
"""


import datetime as dt
import os

import click
import toml
from loguru import logger

from pypsse.common import SIMULATION_SETTINGS_FILENAME
from pypsse.profile_manager.common import (
    DEFAULT_PROFILE_NAME,
    DEFAULT_PROFILE_RESOLUTION,
    DEFAULT_PROFILE_TYPE,
    DEFAULT_START_TIME,
    PROFILE_VALIDATION,
)
from pypsse.profile_manager.profile_store import ProfileManager


@click.argument(
    "project-path",
)
@click.option(
    "-f",
    "--csv-file-path",
    help="Path to a csv valid file",
    required=False,
    default="",
)
@click.option(
    "-p",
    "--profile-folder",
    help="""Path to folder containing csv profiles.
    CSV file names should follow the following format: {profile-type}_{profile-name}""",
    required=False,
    default="",
)
@click.option(
    "-n",
    "--profile-name",
    required=False,
    default=DEFAULT_PROFILE_NAME,
    show_default=True,
    help="Profile name",
)
@click.option(
    "-t",
    "--profile-type",
    required=False,
    default=DEFAULT_PROFILE_TYPE,
    show_default=True,
    help=f"Profile type; Possible values: {list(PROFILE_VALIDATION.keys())}",
)
@click.option(
    "-T",
    "--start-time",
    required=False,
    default=DEFAULT_START_TIME,
    show_default=True,
    help="Time index for the first time step, format = Y-m-d H:M:S.f",
)
@click.option(
    "-r",
    "--profile-res",
    required=False,
    default=DEFAULT_PROFILE_RESOLUTION,
    show_default=True,
    help="Profile time resolution in seconds",
)
@click.option(
    "-i",
    "--profile-info",
    required=False,
    default="",
    show_default=True,
    help="Profile time resolution in seconds",
)
@click.command()
def create_profiles(
    project_path, csv_file_path, profile_folder, profile_name, profile_type, start_time, profile_res, profile_info
):
    """Creates profiles for PyPSSE project."""
    settings_file = os.path.join(project_path, "Settings", SIMULATION_SETTINGS_FILENAME)
    if os.path.exists(settings_file):
        if csv_file_path and os.path.exists(csv_file_path):
            settings = toml.load(settings_file)
            a = ProfileManager(None, settings)
            a.add_profiles_from_csv(
                csv_file=csv_file_path,
                name=profile_name,
                pType=profile_type,
                startTime=dt.datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S.%f").astimezone(None),
                resolution_sec=profile_res,
                info=profile_info,
            )
            logger.info(f"Profile '{profile_name}' added to group '{profile_type}'")
        elif os.path.exists(profile_folder):
            settings = toml.load(settings_file)
            a = ProfileManager(None, settings)
            for _, _, files in os.walk(profile_folder):
                for file in files:
                    if file.endswith(".csv"):
                        filename = file.replace(".csv", "")
                        if "__" in filename:
                            dtype, p_name = filename.split("__")
                            a.add_profiles_from_csv(
                                csv_file=os.path.join(profile_folder, file),
                                name=p_name,
                                pType=dtype,
                                startTime=dt.datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S.%f").astimezone(None),
                                resolution_sec=profile_res,
                                info=profile_info,
                            )
                            msg = f"Profile '{p_name}'' added to group '{dtype}'"
                            logger.info(msg)
        else:
            msg = "Value for either -f or -p flag has to be passed"
            raise Exception(msg)
    else:
        msg = f"{project_path} is not a valid pypsse project"
        raise Exception(msg)



# Handle complex and nominal this way
"Constant definations for the simulation modes"

import copy

dyn_only_options = {
    "Loads": {
        "lmodind": {
            "TD": 14,
            "TC": 15,
            "FmA": 18,
            "FmB": 19,
            "FmC": 20,
            "FmD": 21,
            "Fel": 22,
            "PFel": 23,
        }
    },
}

DYNAMIC_ONLY_PPTY = {
    "Loads": {
        "FmA": ["FmA"],
        "FmB": ["FmB"],
        "FmC": ["FmC"],
        "FmD": ["FmD"],
        "Fel": ["Fel"],
        "PFel": ["PFel"],
        "TD": ["TD"],
        "TC": ["TC"],
    }
}

STANDARD_FORMAT = {
    "Buses": {
        "Number": ["NUMBER"],
        "BusNum": ["NUMBER"],
        "NomkV": ["BASE"],
        "kV": ["KV"],
        "Vpu": ["PU"],
        "Vangle": ["ANGLE"],  # not sure whether degree or radian is expected in NAERM
        "AreaNumber": ["AREA"],
        "ZoneNumber": ["ZONE"],
        "AreaName": ["AREANAME"],
        "SubLongitude": ["LONG"],
        "SubLatitude": ["LATI"],
        "LoadMW": ["TOTAL", "REAL"],
        "LoadMvar": ["TOTAL", "IMAG"],
        "NomB": ["YS", "IMAG", "NOM"],
        "NomG": ["YS", "REAL", "NOM"],
        "GenMW": ["GENPOWER", "REAL"],
        "GenMvar": ["GENPOWER", "IMAG"],
        "SubNumber": ["STATION"],
        "ShuntMVAR": ["YS", "IMAG", "ACT"],
        "Name": ["NAME"],
        "Status": ["STATUS"],
        "IsFeederEligible": ["ISLOADBUS"],
        "BusName": ["NAME"],
        "BusNomVolt": ["BASE"],
        "BusPUVolt": ["PU"],
    },
    "Branches": {
        "LimitMVAA": ["RATEA"],
        "LimitMVAB": ["RATEB"],
        "LimitMVAC": ["RATEC"],
        "Status": ["STATUS"],
        "B": ["BY"],  # BY for susceptance using Star model
        "BusNameFrom": ["FROMBUSNAME"],
        "BusNumFrom": ["FROMBUSNUM"],
        "BusNameTo": ["TOBUSNAME"],
        "BusNumTo": ["TOBUSNUM"],
        "Circuit": ["CIRCUIT"],
        "SubNumberTo": ["SUBNUMBERTO"],
        "SubNumberFrom": ["SUBNUMBERFROM"],
        "BusNum": ["FROMBUSNUM"],
        "BusNum:1": ["TOBUSNUM"],
        "LineCircuit": ["CIRCUIT"],
        "LineStatus": ["STATUS"],
        "LineMaxPercentAmp": ["PCTRTA"],
        "NomkVFrom": ["NOMKVFROM"],
        "NomKVTo": ["NOMKVTO"],
        "R": ["RX", "REAL"],
        "X": ["RX", "IMAG"],
    },
    "Machines": {
        "BusName": ["BUSNAME"],
        "BusNum": ["BUSNUM"],
        "ID": ["MACID"],
        "MWMin": ["PMIN"],
        "MWMax": ["PMAX"],
        "MvarMin": ["QMIN"],
        "MvarMax": ["QMAX"],
        "MW": ["P"],
        "Mvar": ["Q"],
        "Status": ["STATUS"],
        "SubNumber": ["SUBNUMBER"],
        "SubLatitude": ["SUBLATITUDE"],
        "SubLongitude": ["SUBLONGITUDE"],
        "MachineID": ["MACID"],
        "AreaNumber": ["AREANUMBER"],
    },
    "Stations": {
        "SubName": ["SUBNAME"],
        "SubNum": ["SubNum"],
        "Buses": ["BUSES"],
        "Gens": ["GENERATORS"],
        "Trans": ["TRANSFORMERS"],
        "NomkV": ["NOMKV"],
        "LoadMW": ["LOADMW"],
        "GenMW": ["GENMW"],
    },
    "Areas": {"TotalGenMW": ["GEN", "REAL"], "TotalGenMvar": ["GEN", "IMAG"], "AreaNum": ["AREANUMBER"]},
    "Zones": {"TotalGenMW": ["GEN", "REAL"], "TotalGenMvar": ["GEN", "IMAG"], "ZoneName": ["ZONENAME"]},
    "Fixed_shunts": {"ShuntMW": ["ACT", "REAL"], "ShuntMvar": ["ACT", "IMAG"], "FXShuntID": ["FXSHID"]},
    "Loads": {
        "LoadMW": ["MVA", "REAL"],
        "LoadID": ["LOADID"],
        "MW": ["TOTAL", "REAL"],
        "Mvar": ["TOTAL", "IMAG"],
        "SMW": ["MVA", "REAL"],
        "SMvar": ["MVA", "IMAG"],
        "ZMW": ["YL", "REAL"],
        "ZMvar": ["YL", "IMAG"],
        "IMW": ["IL", "REAL"],
        "IMvar": ["IL", "IMAG"],
        "ID": ["LOADID"],
        "BusNum": ["BUSNUM"],
        "Status": ["STATUS"],
    },
    "Induction_generators": {"IndID": ["INDID"]},
    "Switched_shunts": {"BusNum": ["BUSNUM"]},
    "Transformers": {
        "FromBus": ["FROMBUSNUM_3WDG"],
        "ToBus": ["TOBUSNUM_3WDG"],
        "ToBus2": ["TOBUS2NUM_3WDG"],
        "TransCircuit": ["CIRCUIT_3WDG"],
    },
    "DCtransmissionlines": {"BusNum": ["RECT"], "BusNum:1": ["INV"], "DCLID": ["DCLINENAME"]},
}


def converter(func):
    def wrapper(*args, **kwargs):
        new_args = list(args)

        def return_data_type(data_list, cls_type):
            for data in data_list:
                if isinstance(data, cls_type):
                    return data

        if "quantities" in kwargs:
            quantities = kwargs["quantities"]
        else:
            quantities = return_data_type(new_args, dict)

        if "subsystem_buses" in kwargs:
            buses = kwargs["subsystem_buses"]
        else:
            buses = return_data_type(new_args, list)

        new_args = [new_args[0], quantities, buses]

        """ Map NAERM keys to PyPSSE keys """
        ext_string2_info = {}
        complex_conversion_dict = {}
        mapping_dict = copy.deepcopy(quantities)
        for class_name, var_list in quantities.items():
            if class_name in STANDARD_FORMAT:
                new_vars = copy.deepcopy(var_list)
                for v in var_list:
                    if v in STANDARD_FORMAT[class_name]:
                        naerm_element_array = STANDARD_FORMAT[class_name][v]
                        new_vars = [naerm_element_array[0] if el == v else el for el in new_vars]

                        if len(naerm_element_array) == 3:  # noqa: PLR2004
                            ext_string2_info[naerm_element_array[0]] = naerm_element_array[2]
                        if len(naerm_element_array) > 1:
                            if class_name not in complex_conversion_dict:
                                complex_conversion_dict[class_name] = {}
                            complex_conversion_dict[class_name][v] = naerm_element_array[1]
                quantities[class_name] = new_vars

        """ Pass extra string info if a variable requires it """

        if "ext_string2_info" not in kwargs:
            kwargs["ext_string2_info"] = {}
        elif kwargs["ext_string2_info"] is None:
            kwargs["ext_string2_info"] = {}

        kwargs["ext_string2_info"].update(ext_string2_info)
        kwargs["mapping_dict"] = mapping_dict

        result_dict = func(*args, **kwargs)

        """ Convert complex values """

        for class_name_, sub_dict in result_dict.items():
            class_name, param = class_name_.split("_")[0], class_name_.split("_")[1]
            class_name_with_ = [keys for keys in STANDARD_FORMAT.keys() if "_" in keys and keys in class_name_]
            if len(class_name_with_) != 0:
                class_name = class_name_with_[0]
                param = class_name_.split("_")[-1]

            # convert status to string
            status_translation = {1: "connected", 0: "notconnected"}
            if param.lower() == "status":
                for key, value in sub_dict.items():
                    sub_dict[key] = status_translation[value]

            if class_name in complex_conversion_dict:
                conv_dict = complex_conversion_dict[class_name]
                if param in conv_dict:
                    for key, value in sub_dict.items():
                        if isinstance(value, complex):
                            if conv_dict[param] == "REAL":
                                sub_dict[key] = value.real
                            elif conv_dict[param] == "IMAG":
                                sub_dict[key] = value.imag

        return result_dict

    return wrapper



from enum import IntEnum

PROFILE_VALIDATION = {
    "Load": ["PL", "QL", "IP", "IQ", "YP", "YQ", "PG", "QG"],
    "Induction_machine": [
        "MBASE",
        "RATEKVPSET",
        "H",
        "A",
        "B",
        "D",
        "E",
        "RA",
        "XA",
        "XM",
        "R1",
        "X1",
        "R2",
        "X2",
        "X3",
        "E1",
        "SE1",
        "E2",
        "SE2",
        "IA1",
        "IA2",
        "XAMULT",
    ],
    "Machine": [
        "PG",
        "QG",
        "QT",
        "QB",
        "PT",
        "PB",
        "MBASE",
        "ZR",
        "ZX",
        "RT",
        "XT",
        "GTAP",
        "F1",
        "F2",
        "F3",
        "F4",
        "WPF",
    ],
    "Plant": ["VS", "RMPCT"],
}

DEFAULT_PROFILE_NAME = "Default"
DEFAULT_START_TIME = "2020-01-01 00:00:00.00"
DEFAULT_PROFILE_TYPE = "Load"
DEFAULT_PROFILE_RESOLUTION = 1.0


class ProfileTypes(IntEnum):
    Load = 0
    Induction_machine = 1
    Machine = 2
    Plant = 3

    @staticmethod
    def names():
        return [c.name for c in ProfileTypes]
        # list(map(lambda c: c.name, ProfileTypes))

    @staticmethod
    def values():
        return [c.value for c in ProfileTypes]
